//-----------------------------------------------------------------------------
// Copyright  2009 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

__FMACRO_BEGIN__
//-----------------------------------------------------------------------------
// macro	mat_inv_2x2
//
// purpose	Compute the inverse of a non-singular 2x2 matrix.
//
// arguments	1 (number[2][2]) array where to store inverted matrix
//		2 (number[2][2]) array of the matrix to invert
//
// yields	(number) determinate (if zero, inverse is not computed)
//
// note		Because this is a macro, types float, double, and long double
//		may be used as the basis type for each matrix array, including
//		mixing different types.
//
// warning	These matrices must not overlap in memory.
//-----------------------------------------------------------------------------
#define mat_inv_2x2(i,m) ({							\
	__typeof__(**(m)) __dt__;						\
	__dt__ = (m)[0][0]*(m)[1][1] - (m)[0][1]*(m)[1][0];			\
	if ( __dt__ != 0.0 ) {							\
		(i)[0][0] =  (m)[1][1] / __dt__;				\
		(i)[0][1] = -(m)[0][1] / __dt__;				\
		(i)[1][0] = -(m)[1][0] / __dt__;				\
		(i)[1][1] =  (m)[0][0] / __dt__;				\
	}									\
	__dt__;									\
})

//-----------------------------------------------------------------------------
// macro	mat_inv_2x3
//
// purpose	Compute the inverse of an extended 2x3 matrix.  Normally a
//		matrix must be square to compute an inverse.  For the 2x3
//		matrix, we add a virtual 3rd row of [0 0 1] and invert this
//		like a 3x3 matrix.  The 3rd row remains the same when it is
//		initially [0 0 1], so it need not be present in the matrix.
//		This simplifies the calculation over a true 3x3 matrix.
//
//		The expected purpose for a 2x3 matrix, and this "inverse" is
//		performing an affine transformation on an object by means of
//		applying the calculation to the coordinate system instead.
//		The coordinate system (point being considered) is multiplied
//		by the inverse and then compared to the object coordinates.
//
// arguments	1 (number[2][3]) array where to store inverted matrix
//		2 (number[2][3]) array of the matrix to invert
//
// yields	(number) determinate (if zero, inverse is not computed)
//
// note		Because this is a macro, types float, double, and long double
//		may be used as the basis type for each matrix array, including
//		mixing different types.
//
// warning	These matrices must not overlap in memory.
//-----------------------------------------------------------------------------
#define mat_inv_2x3(i,m) ({							\
	__typeof__(**(m)) __dt__;						\
	__dt__ = (m)[0][0]*(m)[1][1] - (m)[0][1]*(m)[1][0];			\
	if ( __dt__ != 0.0 ) {							\
		(i)[0][0] =  (m)[1][1] / __dt__;				\
		(i)[0][1] = -(m)[0][1] / __dt__;				\
		(i)[0][2] = ((m)[0][1]*(m)[1][2]-(m)[0][2]*(m)[1][1]) / __dt__;	\
		(i)[1][0] = -(m)[1][0] / __dt__;				\
		(i)[1][1] =  (m)[0][0] / __dt__;				\
		(i)[1][2] = ((m)[0][2]*(m)[1][0]-(m)[0][0]*(m)[1][2]) / __dt__;	\
	}									\
	__dt__;									\
})

//-----------------------------------------------------------------------------
// macro	mat_inv_3x3
//
// purpose	Compute the inverse of a non-singular 3x3 matrix.
//
// arguments	1 (number[3][3]) array where to store inverted matrix
//		2 (number[3][3]) array of the matrix to invert
//
// yields	(number) determinate (if zero, inverse is not computed)
//
// note		Because this is a macro, types float, double, and long double
//		may be used as the basis type for each matrix array, including
//		mixing different types.
//
// warning	These matrices must not overlap in memory.
//
// note		The destination is used for intermediate calculations
//		and will be modified even if the matrix is singular.
//		The caller must not depend on previous values in the
//		inverse matrix space being unchanged when the matrix
//		being inverted is singular.
//-----------------------------------------------------------------------------
#define mat_inv_3x3(i,m) ({							\
	__typeof__(**(m)) __dt__;						\
	(i)[0][0] = (m)[1][1]*(m)[2][2] - (m)[1][2]*(m)[2][1];			\
	(i)[0][1] = (m)[0][2]*(m)[2][1] - (m)[0][1]*(m)[2][2];			\
	(i)[0][2] = (m)[0][1]*(m)[1][2] - (m)[0][2]*(m)[1][1];			\
	__dt__ = (m)[0][0]*(i)[0][0]+(m)[1][0]*(i)[0][1]+(m)[2][0]*(i)[0][2];	\
	if ( __dt__ != 0.0 ) {							\
		(i)[0][0] /= __dt__;						\
		(i)[0][1] /= __dt__;						\
		(i)[0][2] /= __dt__;						\
		(i)[1][0] = ((m)[1][2]*(m)[2][0]-(m)[1][0]*(m)[2][2]) / __dt__;	\
		(i)[1][1] = ((m)[0][0]*(m)[2][2]-(m)[0][2]*(m)[2][0]) / __dt__;	\
		(i)[1][2] = ((m)[0][2]*(m)[1][0]-(m)[0][0]*(m)[1][2]) / __dt__;	\
		(i)[2][0] = ((m)[1][0]*(m)[2][1]-(m)[1][1]*(m)[2][0]) / __dt__;	\
		(i)[2][1] = ((m)[0][1]*(m)[2][0]-(m)[0][0]*(m)[2][1]) / __dt__;	\
		(i)[2][2] = ((m)[0][0]*(m)[1][1]-(m)[0][1]*(m)[1][0]) / __dt__;	\
	}									\
	__dt__;									\
})

__FMACRO_END__
