//-----------------------------------------------------------------------------
// Copyright  2007 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/helloworld
// homepage	http://phil.ipal.org/freeware/libh/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include <libh_common.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <utime.h>
#include <unistd.h>

#include <libh/ftr.h>
#include <libh/io.h>
#include <libh/string.h>


//-----------------------------------------------------------------------------
// program	setdirtime
//
// purpose	Set all directories in a file tree to have a modified date of
//		the latest modified date of all objects within.  An empty
//		directory is not set.
//
// syntax	setdirtime  [ directoryname ... ]
//
// returns	(int) >= 0 : success
//		(int) == EOF : error
//-----------------------------------------------------------------------------
int
main (
    int		argc
    ,
    char * *	argv
    ,
    char * *	envp
    )
{
    struct timestack {
	time_t			mtime		;
	struct timestack *	next		;
	struct timestack *	prev		;
    };

    struct timeval		times	[2]	;
    struct timestack *		time_stack	;
    FTR				this_ftr	;
    struct timestack		time_root	;
    const char *		dir_name	;
    const char *		pgm_name	;
    int				file_type	;


    dir_name = NULL;
    time_stack = & time_root;
    time_stack->mtime = 0;
    time_stack->next = NULL;
    time_stack->prev = NULL;

    pgm_name = str_tail_one( argv[0], '/' );

    //---------------------------------------------------------------
    // Iterate through all names given, or once with name being NULL.
    //---------------------------------------------------------------
    if ( argc > 1 ) dir_name = argv[1];
    do {
	this_ftr = ftr_new( dir_name );
	if ( ! this_ftr ) {
	    if ( dir_name ) {
		fprintf( stderr, "%s: unable to access \"%s\": %s\n", pgm_name, dir_name, strerror( errno ) );
	    } else {
		fprintf( stderr, "%s: unable to access current directory: %s\n", pgm_name, strerror( errno ) );
	    }
	    return 1;
	}

	//-----------------------------------------------
	// Iterate through every object in the file tree.
	//-----------------------------------------------
	while ( ( file_type = ftr_get_parent( this_ftr ) ) > 0 ) {
	    switch ( file_type ) {

	    //-----------------------------
	    // Descend into this directory.
	    //-----------------------------
	    case 'd':
		//-- If the time list needs to be extended, allocate and link in a new node.
		if ( ! time_stack->next ) {
		    time_stack->next = malloc( sizeof (struct timestack) );
		    if ( ! time_stack->next ) {
			fprintf( stderr, "%s: malloc( %lu ): out of memory\n",
				 pgm_name, (unsigned long) ( sizeof (struct timestack) ) );
		    }
		    time_stack->next->next = NULL;
		    time_stack->next->prev = time_stack;
		}

		//-- Switch to the next node in the time list.
		time_stack = time_stack->next;
		time_stack->mtime = 0;
		break;

	    //--------------------------------------------------------
	    // This directory is done, so set its mtime, if different.
	    //--------------------------------------------------------
	    case 'a':
		if ( time_stack->mtime != 0 && time_stack->mtime != ftr_stat( this_ftr ).st_mtime ) {
		    ftr_stat( this_ftr ).st_mtime = time_stack->mtime;
		    times[1].tv_sec  = time_stack->mtime;
		    times[0].tv_sec  = ftr_stat( this_ftr ).st_atime;
		    times[1].tv_usec = 0;
		    times[0].tv_usec = 0;
		    if ( utimes( ftr_name_last( this_ftr ), times ) < 0 ) {
			fprintf( stderr, "%s: Error %u (%s) setting mtime for \"%s\"\n",
				 pgm_name, errno, strerror( errno ), ftr_name_full( this_ftr ) );
		    }
		}

		//--------------------------------------
		// Back up while keeping the list space.
		//--------------------------------------
		time_stack = time_stack->prev;
		if ( ! time_stack ) {
		    fprintf( stderr,
			     "Inconsistent file tree ascends above starting directory:\n"
			     "    %s\n"
			     "The file tree may have changed during recursion.\n",
			     dir_name );
		}

		//-----------------------------------------------------
		// Fall through to test the ascended directory's mtime.
		//-----------------------------------------------------

	    //-------------------------------
	    // Keep whichever mtime is later.
	    //-------------------------------
	    case 'b':
	    case 'c':
	    case 'f':
	    case 'p':
	    case 's':
		if ( time_stack->mtime < ftr_stat( this_ftr ).st_mtime ) {
		    time_stack->mtime = ftr_stat( this_ftr ).st_mtime;
		}
		break;

	    //-----------------------------------------------------
	    // Ignore symlinks as their times are meaningless here.
	    //-----------------------------------------------------
	    case 'l':
		break;

	    //------------------------------------
	    // Handle recognized error conditions.
	    //------------------------------------
	    case '-':
		fprintf( stderr, "%s: no status available for \"%s\"\n",
			 pgm_name, ftr_name_full( this_ftr ) );
		break;

	    case '?':
		fprintf( stderr, "%s: unknown filetype for \"%s\"\n",
			 pgm_name, ftr_name_full( this_ftr ) );
		break;

	    case 'e':
		fprintf( stderr, "%s: unable to descend into directory \"%s\": read error\n",
			 pgm_name, ftr_name_full( this_ftr ) );
		break;

	    case 'm':
		fprintf( stderr, "%s: unable to descend into directory \"%s\": limit exceeded\n",
			 pgm_name, ftr_name_full( this_ftr ) );
		break;

	    //------------------------------------
	    // Handle everything else as an error.
	    //------------------------------------
	    default:
		fprintf( stderr, "%s: unrecognized filetype from ftr_get(): '%c'\n",
			 pgm_name, file_type );
		break;

	    }
	}

	ftr_end( this_ftr );

    } while ( -- argc > 0 && ( dir_name = * ++ argv ) );

    //---------------------------------------------------------------
    // We could clean up the allocated time list, but why bother now.
    //---------------------------------------------------------------
    return 0;
}
