//-----------------------------------------------------------------------------
// Copyright  2003 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/io
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <unistd.h>

#include "io_lib.h"

extern char **environ;

__PROTO_BEGIN__
//-----------------------------------------------------------------------------
// function	open_fork_execv
//
// purpose	Fork a child process with one or more pipes between parent and
//		child which will be stdin, stdout, and/or stderr for child.
//
// arguments	1 (int *) where to store file descriptor of write end of pipe
//			which will be the child's stdin.
//		2 (int *) where to store file descriptor of read end of pipe
//			which will be the child's stdout.
//		3 (int *) where to store file descriptor of read end of pipe
//			which will be the child's stderr.
//		4 (const char *) filename of program to execute in child
//		5 (char * const *) arguments to program being executed
//		6 (char * const *) environment variables to pass to program
//			or NULL to pass current environment
//
// returns	(int) -1 if there is an error (see errno)
//		(int) child PID
//
// note		If a pointer has the value NULL, then no pipe will be created
//		for that particular purpose, and the child will inherit the
//		same file descriptor as the parent.
//
// note		If the pointers given in arguments 2 and 3 are the same, then
//		only a single pipe will be created for child output, and that
//		pipe will be duplicated for both stdout and stderr.
//
// warning	Care should be taken when transferring data both to and from
//		a child process at the same time to ensure that the parent and
//		child do not get into deadlock condition waiting for buffered
//		data from each other.  This is especially important when the
//		child process executes another program which may not handle
//		buffering as expected (such as a data compression program).
//-----------------------------------------------------------------------------
int
open_fork_execv (
    int *		arg_fd_ptr_stdin
    ,
    int *		arg_fd_ptr_stdout
    ,
    int *		arg_fd_ptr_stderr
    ,
    const char *	arg_execute
    ,
    char * const *	arg_argv
    ,
    char * const *	arg_envp
    )
__PROTO_END__
{
    int			pipe_stdin	[2];
    int			pipe_stdout	[2];
    int			pipe_stderr	[2];
    int			fork_pid	;


    //---------------------------------------
    // Initialize pipe descriptors as closed.
    //---------------------------------------
    pipe_stdin[0]  = pipe_stdin[1]  = -1;
    pipe_stdout[0] = pipe_stdout[1] = -1;
    pipe_stderr[0] = pipe_stderr[1] = -1;

    //--------------------
    // Create the pipe(s).
    //--------------------
    if ( arg_fd_ptr_stdin ) {
	if ( pipe( pipe_stdin ) < 0 ) goto error_cleanup_return;
    }
    if ( arg_fd_ptr_stdout ) {
	if ( pipe( pipe_stdout ) < 0 ) goto error_cleanup_return;
    }
    if ( arg_fd_ptr_stderr ) {
	if ( arg_fd_ptr_stderr == arg_fd_ptr_stdout ) {
	    pipe_stderr[0] = pipe_stdout[0];
	    pipe_stderr[1] = pipe_stdout[1];
	} else {
	    if ( pipe( pipe_stderr ) < 0 ) goto error_cleanup_return;
	}
    }

    //------------------------------
    // Flush standard library files.
    //------------------------------
    fflush( stdout );
    fflush( stderr );

    //--------------------------
    // Create the child process.
    //--------------------------
    fork_pid = fork();
    if ( fork_pid < 0 ) goto error_cleanup_return;

    //----------------------------------
    // Handle child setup and execution.
    //----------------------------------
    if ( fork_pid == 0 ) {

	//------------------------------------------------
	// Get the stdin descriptor into proper alignment.
	//------------------------------------------------
	if ( arg_fd_ptr_stdin ) {
	    close( pipe_stdin[1] );
	    if ( pipe_stdout[1] == STDIN_FILENO ) {
		if ( ( pipe_stdout[1] = dup( STDIN_FILENO ) ) < 0 ) _exit( 1 );
	    }
	    if ( pipe_stderr[1] == STDIN_FILENO ) {
		if ( ( pipe_stderr[1] = dup( STDIN_FILENO ) ) < 0 ) _exit( 1 );
	    }
	    if ( pipe_stdin[0] != STDIN_FILENO ) {
		if ( dup2( pipe_stdin[0], STDIN_FILENO ) < 0 ) _exit( 1 );
		close( pipe_stdin[0] );
	    }
	}

	//-------------------------------------------------
	// Get the stdout descriptor into proper alignment.
	//-------------------------------------------------
	if ( arg_fd_ptr_stdout ) {
	    close( pipe_stdout[0] );
	    if ( pipe_stderr[1] == STDOUT_FILENO ) {
		if ( ( pipe_stderr[1] = dup( STDOUT_FILENO ) ) < 0 ) _exit( 1 );
	    }
	    if ( pipe_stdout[1] != STDOUT_FILENO ) {
		if ( dup2( pipe_stdout[1], STDOUT_FILENO ) < 0 ) _exit( 1 );
		close( pipe_stdout[1] );
	    }
	}

	//-------------------------------------------------
	// Get the stderr descriptor into proper alignment.
	//-------------------------------------------------
	if ( arg_fd_ptr_stderr ) {
	    close( pipe_stderr[0] );
	    if ( pipe_stderr[1] != STDERR_FILENO ) {
		if ( dup2( pipe_stderr[1], STDERR_FILENO ) < 0 ) _exit( 1 );
		close( pipe_stderr[1] );
	    }
	}

	//------------------------------------------
	// Execute specified program with arguments.
	//------------------------------------------
	execve( arg_execute, arg_argv, arg_envp ? arg_envp : environ );
	_exit( 1 );
    }

    //------------------------------------------
    // The parent must not have child ends open.
    //------------------------------------------
    if ( pipe_stdin[0]  >= 0 ) close( pipe_stdin[0]  );
    if ( pipe_stdout[1] >= 0 ) close( pipe_stdout[1] );
    if ( pipe_stderr[1] >= 0 ) close( pipe_stderr[1] );

    //-----------------------------------
    // Store file descriptors for caller.
    //-----------------------------------
    if ( arg_fd_ptr_stdin  ) {
	* arg_fd_ptr_stdin  = pipe_stdin[1];
    }
    if ( arg_fd_ptr_stdout ) {
	* arg_fd_ptr_stdout = pipe_stdout[0];
    }
    if ( arg_fd_ptr_stderr ) {
	* arg_fd_ptr_stderr = ( arg_fd_ptr_stderr == arg_fd_ptr_stdout ) ? pipe_stdout[0] : pipe_stderr[0];
    }

    //-----------------------------------------------------
    // Return child process ID to caller in parent process.
    //-----------------------------------------------------
    return fork_pid;

    //----------------------------------------------
    // Do error cleanup and return error indication.
    //----------------------------------------------
 error_cleanup_return:
    if ( pipe_stderr[0] >= 0 )	close( pipe_stderr[0] );
    if ( pipe_stderr[1] >= 0 )	close( pipe_stderr[1] );
    if ( pipe_stdout[0] >= 0 )	close( pipe_stdout[0] );
    if ( pipe_stdout[1] >= 0 )	close( pipe_stdout[1] );
    if ( pipe_stdin[0]  >= 0 )	close( pipe_stdin[0]  );
    if ( pipe_stdin[1]  >= 0 )	close( pipe_stdin[1]  );
    return -1;
}
