//-----------------------------------------------------------------------------
// Copyright  2003 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/io
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include "io_lib.h"

__FMACRO_BEGIN__
//-----------------------------------------------------------------------------
// macro	readlink_join
//
// purpose	Read the reference data for symbolic link of the given name
//		where the symbolic link name is given as multiple string
//		arguments to be joined together.
//
// arguments	1 (char *) target buffer
//		2 (size_t) size of target buffer
//		3..N (const char *) symbolic link name parts
//
// returns	(ssize_t)  < 0 : [failure, see errno]
//		(ssize_t) >= 0 : [success] number of characters stored
//
// note		Unlike the POSIX/SUS standard readlink() function, this
//		function/macro will properly terminate the symlink data.
//		Thus it will be necessary to provide at least one additional
//		character of buffer space.
//-----------------------------------------------------------------------------
#define readlink_join(b,s,n...) ((readlink_join)((b),(s),n,(const char*)(NULL)))

__FMACRO_END__

__PROTO_BEGIN__
//-----------------------------------------------------------------------------
// function	(readlink_join)
//
// purpose	Read the reference data for symbolic link of the given name
//		where the symbolic link name is given as multiple string
//		arguments to be joined together.
//
// arguments	1 (char *) target buffer
//		2 (size_t) size of target buffer
//		3..N (const char *) symbolic link name parts
//		N+1 (const char *) NULL to indicate end of list
//
// returns	(ssize_t)  < 0 : [failure, see errno]
//		(ssize_t) >= 0 : [success] number of characters stored
//
// note		Unlike the POSIX/SUS standard readlink() function, this
//		function/macro will properly terminate the symlink data.
//		Thus it will be necessary to provide at least one additional
//		character of buffer space.
//-----------------------------------------------------------------------------
ssize_t
(readlink_join) (
    char *		arg_buffer
    ,
    size_t		arg_size
    ,
    const char *	arg_name
    ,
    ...
    )
__PROTO_END__
{
    va_list		var_args	;
    const char *	str		;
    char *		name		;
    char *		ptr		;
    size_t		len		;
    ssize_t		result		;


    //--------------------------------------------------------------
    // First figure out how much space is needed for the given name.
    //--------------------------------------------------------------
    len = 1;
    va_start( var_args, arg_name );
    str = arg_name;
    while ( str ) {
	len += strlen( str );
	str = va_arg( var_args, const char * );
    }
    va_end ( var_args );

    //---------------------------------------------
    // Temporarily allocate the exact space needed.
    //---------------------------------------------
    if ( ! ( name = alloca( len ) ) ) return -1;

    //--------------------
    // Construct the name.
    //--------------------
    ptr = name;
    va_start( var_args, arg_name );
    while ( str ) {
	while ( * str ) * ptr ++ = * str ++;
	str = va_arg( var_args, const char * );
    }
    va_end ( var_args );
    * ptr = 0;

    //-----------------------------------------------------------
    // Do the system call function to get the symbolic link data.
    //-----------------------------------------------------------
    result = readlink( name, arg_buffer, arg_size );
    if ( result >= 0 ) {
	if ( result >= arg_size ) result = arg_size - 1;
	arg_buffer[ result ] = 0;
    }
    return result;
}
