//-----------------------------------------------------------------------------
// Copyright  2012 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/keepstr
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include "keepstr_lib.h"

#include <stdio.h>
#include <unistd.h>

//---------------------------------------------------------------------------------------------------
// function	keepstr_cmp (private)
//
// purpose	String compare function for libavl.
//
// arguments	1 (const struct keepstr_s *) pointer to node 1 to access key string
//		2 (const struct keepstr_s *) pointer to node 2 to access key string
//		3 (cost AVL *) pointer to AVL tree root (not needed for keepstr)
//
// returns	(int) state (-1, 0, 1) of string comparison.
//---------------------------------------------------------------------------------------------------
static int
keepstr_cmp (
    struct keepstr_s *	node1_p
    ,
    struct keepstr_s *	node2_p
    ,
    AVL *		tree_p
    )
{
    return strcmp( node1_p->str, node2_p->str );
}

__PROTO_BEGIN__
//---------------------------------------------------------------------------------------------------
// function	keepstr_pl
//
// purpose	Look up the given string.  Add it to the common shared string pool if absent.
//		Return a pointer to the string in the pool.
//
// note		The caller no longer needs to use the original string used for the search.
//
// arguments	1 (const char *) pointer to string to be kept
//		2 (size_t) length of string to be kept
//
// returns	(const char *) == NULL : error (probably out of memory)
//		(const char *) != NULL : pointer to the kept string in the keep map
//---------------------------------------------------------------------------------------------------
// macro	keepstr_pe
//
// purpose	Look up the given string.  Add it to the common shared string pool if absent.
//		Return a pointer to the string in the pool.
//
// note		The caller no longer needs to use the original string used for the search.
//
// arguments	1 (const char *) pointer to string to be kept
//		2 (const char *) pointer to end of string to be kept
//
// returns	(const char *) == NULL : error (probably out of memory)
//		(const char *) != NULL : pointer to the	kept string in the keep	map
//---------------------------------------------------------------------------------------------------
// macro	keepstr
//
// purpose	Look up the given string.  Add it to the common shared string pool if absent.
//		Return a pointer to the string in the pool.
//
// note		The caller no longer needs to use the original string used for the search.
//
// arguments	1 (const char *) pointer to terminated string to be kept
//
// returns	(const char *) == NULL : error (probably out of memory)
//		(const char *) != NULL : pointer to the	kept string in the keep	map
//---------------------------------------------------------------------------------------------------
#define keepstr_pe(p,e) ({							\
	const char *__ptr__;							\
	__ptr__ = (p);								\
	__ptr__ ? (keepstr_pl)( (__ptr__) , ((e)-__ptr__) ) : NULL;		\
})

#define keepstr(s) ({								\
	const char *__ptr__;							\
	__ptr__ = (s);								\
	__ptr__ ? (keepstr_pl)( (__ptr__) , (strlen(__ptr__)) ) : NULL;		\
})

const char *
(keepstr_pl) (
    const char *	arg_ptr
    ,
    size_t		arg_len
    )
__PROTO_END__
{
    static AVL	keepstr_avl	[1]	;
    static int	keepstr_ready	= 0	;
    static char	keepstr_empty	= 0	;

    struct keepstr_s *		node_p	;

    //-- If a NULL pointer is given, give one back.
    if ( ! arg_ptr ) return NULL;

    //-- If an empty string is given, give one back.
    if ( arg_len == 0 ) return & keepstr_empty;

    //-- If not yet initialized, then do so now.
    if ( ! keepstr_ready ) {
	if ( ! avl_init( keepstr_avl, struct keepstr_s, link, keepstr_cmp ) ) return NULL;
	keepstr_ready = 1;
    }

    //-- Build lookup node to see if string is already kept.
    node_p = alloca( arg_len + sizeof (struct keepstr_s) );		if ( ! node_p ) return NULL;
    memcpy( node_p->str, arg_ptr, arg_len );
    node_p->str[arg_len] = 0;
    node_p = avl_find_exact( keepstr_avl, node_p );			if ( node_p ) return node_p->str;

    //-- Build permanent node and keep this string.
    node_p = malloc( arg_len + sizeof (struct keepstr_s) );		if ( ! node_p ) return NULL;
    memcpy( node_p->str, arg_ptr, arg_len );
    avl_insert( keepstr_avl, node_p );

    return node_p->str;
}
