__PREFIX_BEGIN__
//-----------------------------------------------------------------------------
// Copyright  2009 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/list
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// file		list.h
//
// purpose	Define macro functions and statements for list operations.
//-----------------------------------------------------------------------------
#ifndef __LIST_H__
#define __LIST_H__

#include <stddef.h>

__PREFIX_END__

__DEFINE_BEGIN__
//-----------------------------------------------------------------------------
// purpose	Define macro functions and statements for list operations.
//
// description	This list implementation is entirely macro based.  The reason
//		for this is to allow direct operation on programmer defined
//		node structs.  No structs are defined by this implementation.
//		Intsead, the programmer defines these node structs with members
//		named "next" and "prev", which shall be of type pointer to the
//		struct itself.  It it not required for these members to be the
//		first members, or even be in this order.
//
// example	struct my_node {
//		    struct my_node *next;
//		    struct my_node *prev;
//		    ...
//		};
//
// description	The list is formed as a circular linked list.  This means the
//		last node will have forward links back to the first node, and
//		the first node will have reverse links back to the last node.
//		The list is anchored using a pointer to the first node.  This
//		anchor is passed to the macros as an lvalue, instead of as a
//		pointer to the anchor.  This allows an anchor that is a simple
//		variable name to just be passed as that variable name.
//
// iterators	Some macros are provided to be statement iterators.  They are
//		used in place of a for() clause that preceeding a statement
//		or block of code.  A name is provided by the programmer for the
//		iterator variable to be used to access each node in the list.
//		The macro will define this variable.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// macro	list_init
//
// purpose	Initializer a list anchor to the empty state.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	NULL
//-----------------------------------------------------------------------------
#define list_init(anc) ((anc)=(NULL))

//-----------------------------------------------------------------------------
// macro	list_is_empty
//
// purpose	Test if a given list is empty.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(int) 1 if list is empty, 0 if not
//-----------------------------------------------------------------------------
#define list_is_empty(anc) ((NULL)==(anc))

//-----------------------------------------------------------------------------
// macro	list_insert_first
// alias	list_insert_head
//
// purpose	Insert a node into a list at the front, making it become the
//		first node.
//
// arguments	1 (struct *) lvalue of list anchor
//		2 (struct *) pointer to new node to insert
//
// yields	(struct *) pointer to inserted node
//-----------------------------------------------------------------------------
#define list_insert_first(anc,new) ({						\
	__typeof__( (anc) ) * list__anc;					\
	__typeof__( (new) ) list__new;						\
	list__anc = & (anc);							\
	list__new = (new);							\
	if ( * list__anc ) {							\
		list__new->next = ( * list__anc );				\
		list__new->prev = ( * list__anc )->prev;			\
		( * list__anc )->prev->next = list__new;			\
		( * list__anc )->prev = list__new;				\
	} else {								\
		list__new->next = list__new;					\
		list__new->prev = list__new;					\
	}									\
	* list__anc = list__new;						\
	list__new;								\
})
#define list_insert_head list_insert_first

//-----------------------------------------------------------------------------
// macro	list_insert_first_new
// alias	list_insert_head_new
//
// purpose	Allocate a new node and insert it into a list at the front,
//		making it become the first node.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(struct *) pointer to new inserted node (or NULL if failed).
//
// note		The remainder of the new node is not yet initialized even
//		though it is now linked into the list.  The calling program
//		must perform any needed initialization of the node before
//		the node is to be used.
//-----------------------------------------------------------------------------
#define list_insert_first_new(anc) ({						\
	__typeof__( (anc) ) * list__anc;					\
	__typeof__( (anc) ) list__new;						\
	list__anc = & (anc);							\
	list__new = malloc( sizeof ( * list__new ) );				\
	if ( list__new ) {							\
		if ( * list__anc ) {						\
			list__new->next = ( * list__anc );			\
			list__new->prev = ( * list__anc )->prev;		\
			( * list__anc )->prev->next = list__new;		\
			( * list__anc )->prev = list__new;			\
		} else {							\
			list__new->next = list__new;				\
			list__new->prev = list__new;				\
		}								\
		* list__anc = list__new;					\
	}									\
	list__new;								\
})
#define list_insert_head_new list_insert_first_new

//-----------------------------------------------------------------------------
// macro	list_insert_last
// alias	list_insert_tail
//
// purpose	Insert a node into a list at the back, making it become the
//		last node.
//
// arguments	1 (struct *) lvalue of list anchor
//		2 (struct *) pointer to new node to insert
//
// yields	(struct *) pointer to inserted node
//-----------------------------------------------------------------------------
#define list_insert_last(anc,new) ({						\
	__typeof__( (anc) ) * list__anc;					\
	__typeof__( (new) ) list__new;						\
	list__anc = & (anc);							\
	list__new = (new);							\
	if ( * list__anc ) {							\
		list__new->next = ( * list__anc );				\
		list__new->prev = ( * list__anc )->prev;			\
		( * list__anc )->prev->next = list__new;			\
		( * list__anc )->prev = list__new;				\
	} else {								\
		list__new->next = list__new;					\
		list__new->prev = list__new;					\
		* list__anc = list__new;					\
	}									\
	list__new;								\
})
#define list_insert_tail list_insert_last

//-----------------------------------------------------------------------------
// macro	list_insert_last_new
// alias	list_insert_tail_new
//
// purpose	Allocate a new node and insert it into a list at the back,
//		making it become the last node.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(struct *) pointer to new inserted node (or NULL if failed).
//
// note		The remainder of the new node is not yet initialized even
//		though it is now linked into the list.  The calling program
//		must perform any needed initialization of the node before
//		the node is to be used.
//-----------------------------------------------------------------------------
#define list_insert_last_new(anc) ({						\
	__typeof__( (anc) ) * list__anc;					\
	__typeof__( (anc) ) list__new;						\
	list__anc = & (anc);							\
	list__new = malloc( sizeof ( * list__new ) );				\
	if ( list__new ) {							\
		if ( * list__anc ) {						\
			list__new->next = ( * list__anc );			\
			list__new->prev = ( * list__anc )->prev;		\
			( * list__anc )->prev->next = list__new;		\
			( * list__anc )->prev = list__new;			\
		} else {							\
			list__new->next = list__new;				\
			list__new->prev = list__new;				\
			* list__anc = list__new;				\
		}								\
	}									\
	list__new;								\
})
#define list_insert_tail_new list_insert_last_new

//-----------------------------------------------------------------------------
// macro	list_remove_first
// alias	list_remove_head
//
// purpose	Remove the first node from a list.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(struct *) pointer to removed node (or NULL if no node)
//-----------------------------------------------------------------------------
#define list_remove_first(anc) ({						\
	__typeof__( (anc) ) * list__anc;					\
	__typeof__( (anc) ) list__node;						\
	list__anc = & (anc);							\
	list__node = * list__anc;						\
	if ( list__node ) {							\
		* list__anc = list__node->next == list__node			\
			? NULL							\
			: (	list__node->next->prev = list__node->prev,	\
				list__node->prev->next = list__node->next,	\
				list__node->next				\
			);							\
	}									\
	list__node;								\
})
#define list_remove_head list_remove_first

//-----------------------------------------------------------------------------
// macro	list_remove_last
// alias	list_remove_tail
//
// purpose	Remove the last node from a list.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(struct *) pointer to removed node (or NULL if no node)
//-----------------------------------------------------------------------------
#define list_remove_last(anc) ({						\
	__typeof__( (anc) ) * list__anc;					\
	__typeof__( (anc) ) list__node;						\
	list__anc = & (anc);							\
	list__node = * list__anc;						\
	if ( list__node ) {							\
		list__node = list__node->prev;					\
		* list__anc = list__node->next == list__node			\
			? NULL							\
			: (	list__node->next->prev = list__node->prev,	\
				list__node->prev->next = list__node->next,	\
				list__node->next				\
			);							\
	}									\
	list__node;								\
})
#define list_remove_tail list_remove_last

//-----------------------------------------------------------------------------
// macro	list_forward
// alias	list_next
//
// purpose	Given a pointer to a node, yield a pointer to the next node.
//
// arguments	1 (struct *) pointer to a node in a list
//
// yields	(struct *) pointer to next node in list
//-----------------------------------------------------------------------------
#define list_forward(node) ((node)->next)
#define list_next(node) ((node)->next)

//-----------------------------------------------------------------------------
// macro	list_reverse
// alias	list_prev, list_previous
//
// purpose	Given a pointer to a node, yield a pointer to the previous node.
//
// arguments	1 (struct *) pointer to a node in a list
//
// yields	(struct *) pointer to previous node in list
//-----------------------------------------------------------------------------
#define list_reverse(node) ((node)->prev)
#define list_prev(node) ((node)->prev)
#define list_previous(node) ((node)->prev)

//-----------------------------------------------------------------------------
// macro	list_first
// alias	list_head
//
// purpose	Given a list anchor, yield a pointer to the first node.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(struct *) pointer to first node in list (or NULL if no nodes)
//-----------------------------------------------------------------------------
#define list_first(anc) ({							\
	__typeof__( (anc) ) * list__anc;					\
	list__anc = & (anc);							\
	list__anc ? * list__anc : NULL;						\
})
#define list_head list_first

//-----------------------------------------------------------------------------
// macro	list_last
// alias	list_tail
//
// purpose	Given a list anchor, yield a pointer to the last node.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(struct *) pointer to last node in list (or NULL if no nodes)
//-----------------------------------------------------------------------------
#define list_last(anc) ({							\
	__typeof__( (anc) ) * list__anc;					\
	list__anc = & (anc);							\
	list__anc ? ( * list__anc )->prev : NULL;				\
})
#define list_tail list_last

//-----------------------------------------------------------------------------
// macro	list_rotate_forward
// alias	list_rotate
//
// purpose	Given a list anchor, rotate the list forward so the first
//		node becomes the last, and the next node becomes the first.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(struct *) pointer to new first node (NULL if empty)
//-----------------------------------------------------------------------------
#define list_rotate_forward(anc) ({						\
	__typeof__( (anc) ) * list__anc;					\
	list__anc = & (anc);							\
	if ( list__anc ) {							\
		* list__anc = ( * list__anc )->next;				\
	}									\
	* list__anc;								\
})
#define list_rotate list_rotate_forward

//-----------------------------------------------------------------------------
// macro	list_rotate_reverse
//
// purpose	Given a list anchor, rotate the list reverse so the last
//		node becomes the first, and the first node becomes the next.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(struct *) pointer to new first node (NULL if empty)
//-----------------------------------------------------------------------------
#define list_rotate_reverse(anc) ({						\
	__typeof__( (anc) ) * list__anc;					\
	list__anc = & (anc);							\
	if ( list__anc ) {							\
		* list__anc = ( * list__anc )->prev;				\
	}									\
	* list__anc;								\
})

//-----------------------------------------------------------------------------
// macro	list_iterate_forward
// alias	list_iterate
//
// purpose	Given a list anchor and node pointer, iterate from first
//		to last node in the list, for an executed code block.
//
// arguments	1 (struct *) lvalue of list anchor
//		2 (struct *) variable name that will point to each node
//
// warning	Multiple evaluations of arguments will likely occur.
//
// note		This macro creates a statement construction similar to a for
//		or while statement.
//
// usage	list_iterate_forward( this_list, this_node ) { ... }
//-----------------------------------------------------------------------------
#define list_iterate_forward(anc,var) for(__typeof__((anc))var=(anc);var;var=(var->next==(anc)?NULL:var->next))
#define list_iterate list_iterate_forward

//-----------------------------------------------------------------------------
// macro	list_iterate_reverse
//
// purpose	Given a list anchor and node pointer, iterate from last
//		to first node in the list, for an executed code block.
//
// arguments	1 (struct *) lvalue of list anchor
//		2 (struct *) variable name that will point to each node
//
// warning	Multiple evaluations of arguments will likely occur.
//
// note		This macro creates a statement construction similar to a for
//		or while statement.
//
// usage	list_iterate_reverse( this_list, this_node ) { ... }
//-----------------------------------------------------------------------------
#define list_iterate_reverse(anc,var) for(__typeof__((anc))var=((anc)?(var)->prev:NULL);var;var=(var->prev==(anc)->prev?NULL:var->prev))

//-----------------------------------------------------------------------------
// macro	list_flush_forward
// alias	list_flush
//
// purpose	Given a list anchor and node pointer, iterate from first
//		to last node in the list, removing each node as iterated,
//		for an executed code block.
//
// arguments	1 (struct *) lvalue of list anchor
//		2 (struct *) variable name that will point to each node
//
// warning	Multiple evaluations of arguments will likely occur.
//
// note		This macro creates a statement construction similar to a for
//		or while statement.
//
// usage	list_flush_forward( this_list, this_node ) { ... }
//-----------------------------------------------------------------------------
#define list_flush_forward(anc,var) for(__typeof__((anc))var;(var=list_remove_first((anc)));)
#define list_flush list_flush_forward

//-----------------------------------------------------------------------------
// macro	list_flush_reverse
//
// purpose	Given a list anchor and node pointer, iterate from last
//		to first node in the list, removing each node as iterated,
//		for an executed code block.
//
// arguments	1 (struct *) lvalue of list anchor
//		2 (struct *) variable name that will point to each node
//
// warning	Multiple evaluations of arguments will likely occur.
//
// note		This macro creates a statement construction similar to a for
//		or while statement.
//
// usage	list_flush_reverse( this_list, this_node ) { ... }
//-----------------------------------------------------------------------------
#define list_flush_reverse(anc,var) for(__typeof__((anc))var;(var=list_remove_last((anc)));)

//-----------------------------------------------------------------------------
// macro	list_flush_free
//
// purpose	Given a list anchor, free all nodes from the list and make the
//		list anchor have the empty state.  This requires that each node
//		was originally allocated with malloc(), calloc(), realloc(), or
//		another function that allocated from the heap.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(void)
//-----------------------------------------------------------------------------
#define list_flush_free(anc) ({							\
	__typeof__((anc)) * libh__anc;						\
	__typeof__((anc)) libh__first;						\
	__typeof__((anc)) libh__this;						\
	__typeof__((anc)) libh__next;						\
	libh__anc = & (anc);							\
	libh__first = * libh__anc;						\
	libh__this = libh__first;						\
	if ( libh__this ) {							\
		do {								\
			libh__next = libh__this->next;				\
			free( libh__this );					\
			libh__this = libh__next;				\
		while ( libh__this != libh__first );				\
		* libh__anc = NULL;						\
	}									\
})

//-----------------------------------------------------------------------------
// macro	list_count
//
// purpose	Given a list anchor, count the number of nodes in the list.
//
// arguments	1 (struct *) lvalue of list anchor
//
// yields	(unsigned long) count of number of nodes in the list
//-----------------------------------------------------------------------------
#define list_count(anc) ({							\
	__typeof__((anc)) libh__anc;						\
	unsigned long libh__count;						\
	libh__anc = (anc);							\
	libh__count = 0;							\
	list_iterate_forward(libh__anc,libh__node)++libh__count;		\
	libh__count;								\
})

//-----------------------------------------------------------------------------
// macro	list_orient
//
// purpose	Given a list anchor and node pointer, effectively rotate the
//		list so the given node is now the first node in the list.
//
// arguments	1 (struct *) lvalue of list anchor
//		2 (struct *) node pointer (must be in the list)
//
// yields	(struct *) the given node
//-----------------------------------------------------------------------------
#define list_orient(anc,node) ((anc)=(node))

__DEFINE_END__

__SUFFIX_BEGIN__

#endif /* __LIST_H__ */
__SUFFIX_END__
