__PREFIX_BEGIN__
//-----------------------------------------------------------------------------
// Copyright  2012 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/optplus
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// file		libh/optplus.h
//-----------------------------------------------------------------------------
#ifndef __OPTPLUS_H__
#define __OPTPLUS_H__

__PREFIX_END__

__INCLUDE_BEGIN__
#include <stdlib.h>
#include <string.h>

#include <libh/ctools.h>
__INCLUDE_END__

__DEFINE_BEGIN__
//---------------------------------------------------------------------------------------------------
// purpose	Scan an argument list with detection and assistance for options in several forms:
//
//		1.  standalone single prefix ('-' or '+').
//		2.  standalone double prefix ("--" or "-+" or "+-" or "++")
//		3.  switching single letter prefix within an argument ("-a+b" = "-a" "+b")
//		4.  assignment modes for single prefix ("-abc=xyzzy" = "-a" "-b" "-c=xyzzy")
//		5.  assignment modes for double prefix ("--foobar=xyzzy")
//		6.  assignment modes for no prefix ("foobar=xyzzy")
//
//		The API defined here is limited to the defined struct, macros, functions.
//		No global variables are used (thread-safe).  It is a "lite OO" API design.
//
// usage	These functions are used in a different way than classic functions getopt() and
//		getopt_long().	A list of options to be recognized is not provided.  Instead, the
//		caller gets all options and arguments, with option type, name, and maybe a value.
//		The caller then determines if an option is valid or not.  The caller can also get
//		raw arguments for cases where an option takes such an argument.	 Single letter
//		options grouped into one argument are returned one at a time.
//
//		The first step is to call optplus_new() to create a new option scanning object for
//		keeping the context.  These functions are thread-safe only when an option scanning
//		object is NOT shared between threads.  That means different threads can be scanning
//		different option sets concurrently.  Most programs never need to do this since they
//		only scan one command line.
//
//		The main option scanning loop calls optplus_next() with the object pointer obtained
//		from optplus_new().  optplus_next() returns a type that describes failure or success
//		in getting the next option or argument, and the type of option.	 Based on the type
//		of option, the caller can get the option name from optplus_name() and an assignment
//		value from optplus_value() if available.  If the option is a single prefix single
//		character, optplus_char() may also be used to get the option name letter.
//
//		The caller can get plain arguments without any option consideration by using the
//		optplus_arg() function.	 This might typically be done when something indicates that
//		options have ended (such as the first non-option argument, or after a solo option).
//
//		When option scanning is complete, call optplus_end() to clean up.
//
//		Additional functions to enhance processing options include:
//		optplus_name_cmp()		- compare the option name to a string
//		optplus_name_match()		- match the option name to a string
//		optplus_name_match_nocase()	- match the option name to a string ignoring case
//		optplus_name_end()		- get pointer to end of name (after last character)
//		optplus_name_len()		- get length of name string
//		optplus_name_dup()		- get a pointer to a duplicate copy of name
//		optplus_value_end()		- get pointer to end of value (after last character)
//		optplus_value_len()		- get length of value string
//		optplus_value_dup()		- get a pointer to a duplicate copy of value
//		optplus_value_or_arg()		- get the value in the option or next argument
//---------------------------------------------------------------------------------------------------

//---------------------------------------------------------------------------------------------------
// struct	optplus_s
//
// purpose	Hold the state for scanning an argument list for options.
//---------------------------------------------------------------------------------------------------
struct optplus_s {
    const char * const *	arga		; // first argument
    const char * const *	argt		; // this argument
    const char * const *	argx		; // next argument
    const char * const *	argz		; // after last argument
    const char *		char_ptr	; // pointer to single char if single mode
    const char *		name_str	; // pointer to option name
    const char *		value_str	; // pointer to option value
    const char *		name_end	; // end ptr of option name
    const char *		value_end	; // end ptr of option value
    size_t			name_len	; // length of option name
    size_t			value_len	; // length of option value
    long			count		; // count of calls to optplus_next()
    int				type		; // prefix type after optplus_next()
    char			char_code[2]	; // single letter option
    char			pre_char[3]	; // prefix character codes
};
typedef struct optplus_s	optplus_t	;
typedef struct optplus_s *	optplus_p	;
typedef struct optplus_s *	OPTPLUS		;

#define OPTPLUS_TYPE_ERROR		-1
#define OPTPLUS_TYPE_END		0
#define OPTPLUS_TYPE_ARGUMENT		1
#define OPTPLUS_TYPE_MINUS		'-'
#define OPTPLUS_TYPE_PLUS		'+'
#define OPTPLUS_TYPE_MINUS_MINUS	char_code_2('-','-')
#define OPTPLUS_TYPE_MINUS_PLUS		char_code_2('-','+')
#define OPTPLUS_TYPE_PLUS_MINUS		char_code_2('+','-')
#define OPTPLUS_TYPE_PLUS_PLUS		char_code_2('+','+')

//---------------------------------------------------------------------------------------------------
// macro	optplus_arg_adup
// purpose	Get the whole next argument (even if it looks like an option) from an option scanning
//		object as an automatic storage allocated duplicate string.
// WARNING	DO NOT USE THIS MACRO unless you fully understand how the alloca() function works,
//		since this macro uses alloca() to allocate memory.  All caveats about using alloca()
//		also apply to this macro.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(char *) == NULL : error, out of memory
//		(char *) != NULL : pointer to duplicate copy of argument string
//---------------------------------------------------------------------------------------------------
#define optplus_arg_adup(o) ({									\
	optplus_p	___opt___;								\
	const char *	___arg___;								\
	const char *	___end___;								\
	char *		___dup___;								\
	size_t		___len___;								\
	___opt___ = (o);									\
	( ___opt___->argx >= ___optplus___->argx ) ? NULL :					\
	( ! * ___opt___->argx ) ? NULL :							\
	({											\
		___arg___ = * ___opt___->argx;							\
		for ( ___end___ = ___arg___; * ___end___; ++ ___end___ );			\
		___len___ = ___end___ + 1 - ___arg___;						\
		___dup___ = alloca( ___len___ );						\
		if ( ___dup___ ) {								\
			memcpy( ___dup___, ___arg___, ___len___ );				\
			___opt___->argt = ___opt___->argx;					\
			++ ___opt___->argx;							\
		}										\
		___dup___;									\
	})											\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_this_arg
// purpose	Get the original whole argument currently being processed.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(const char *) == NULL : have now reached end of arguments
//		(const char *) != NULL : pointer to current argument
//---------------------------------------------------------------------------------------------------
#define optplus_this_arg(o) ((const char *)((o)->argt))

//---------------------------------------------------------------------------------------------------
// macro	optplus_name
// purpose	Get the latest option name string selected by optplus_next().
// WARNING	This string also includes a value that may be following.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(const char *) == NULL : no name available
//		(const char *) != NULL : pointer to name string
//---------------------------------------------------------------------------------------------------
#define optplus_name(o) ((const char *)(o)->name_str)

//---------------------------------------------------------------------------------------------------
// macro	optplus_name_end
// purpose	Get the end pointer of the latest option name string.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(const char *) == NULL : no name available
//		(const char *) != NULL : pointer to name string
//---------------------------------------------------------------------------------------------------
#define optplus_name_end(o) ((const char *)(o)->name_end)

//---------------------------------------------------------------------------------------------------
// macro	optplus_name_len
// purpose	Get the length of the latest option name string.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(size_t) length of option name string
//---------------------------------------------------------------------------------------------------
#define optplus_name_len(o) ((size_t)((o)->name_len))

//---------------------------------------------------------------------------------------------------
// macro	optplus_name_dup
// purpose	Get the latest option name string as a heap storage allocated duplicate.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(char *) == NULL : error, out of memory
//		(char *) != NULL : pointer to duplicate copy of name string
//---------------------------------------------------------------------------------------------------
#define optplus_name_dup(o) ({									\
	optplus_p ___optplus___;								\
	___optplus___ = (o);									\
	char * ___ptr___;									\
	___ptr___ = malloc( ___optplus___->name_len + 1 );					\
	if ( ___ptr___ ) {									\
		memcpy( ___ptr___, ___optplus___->name_str, ___optplus___->name_len );		\
		___ptr___[ ___optplus___->name_len ] = 0;					\
	}											\
	___ptr___;										\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_name_adup
// purpose	Get the latest option name string as an automatic storage allocated duplicate.
// WARNING	DO NOT USE THIS MACRO unless you fully understand how the alloca() function works,
//		since this macro uses alloca() to allocate memory.  All caveats about using alloca()
//		also apply to this macro.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(char *) == NULL : error, out of memory
//		(char *) != NULL : pointer to duplicate copy of name string
//---------------------------------------------------------------------------------------------------
#define optplus_name_adup(o) ({									\
	optplus_p ___optplus___;								\
	___optplus___ = (o);									\
	char * ___ptr___;									\
	___ptr___ = alloca( ___optplus___->name_len + 1 );					\
	if ( ___ptr___ ) {									\
		memcpy( ___ptr___, ___optplus___->name_str, ___optplus___->name_len );		\
		___ptr___[ ___optplus___->name_len ] = 0;					\
	}											\
	___ptr___;										\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_value
// purpose	Get the latest option value string selected by optplus_next().
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(const char *) == NULL : no value available
//		(const char *) != NULL : pointer to value string
//---------------------------------------------------------------------------------------------------
#define optplus_value(o) ((const char *)((o)->value_str))

//---------------------------------------------------------------------------------------------------
// macro	optplus_value_end
// purpose	Get the end_pointer of the latest option value string.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(const char *) == NULL : no value available
//		(const char *) != NULL : pointer to value string
//---------------------------------------------------------------------------------------------------
#define optplus_value(o) ((const char *)((o)->value_str))

//---------------------------------------------------------------------------------------------------
// macro	optplus_value_len
// purpose	Get the length of the latest option value string.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(size_t) length of option value string
//---------------------------------------------------------------------------------------------------
#define optplus_value_len(o) ((size_t)((o)->value_len))

//---------------------------------------------------------------------------------------------------
// macro	optplus_value_dup
// purpose	Get the latest option value string as a heap storage allocated duplicate.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(char *) == NULL : error, out of memory
//		(char *) != NULL : pointer to duplicate copy of value string
//---------------------------------------------------------------------------------------------------
#define optplus_value_dup(o) ({									\
	optplus_p ___optplus___;								\
	___optplus___ = (o);									\
	char * ___ptr___;									\
	___ptr___ = malloc( ___optplus___->value_len + 1 );					\
	if ( ___ptr___ ) {									\
		memcpy( ___ptr___, ___optplus___->value_str, ___optplus___->value_len );	\
		___ptr___[ ___optplus___->value_len ] = 0;					\
	}											\
	___ptr___;										\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_value_adup
// purpose	Get the latest option value string as an automatic storage allocated duplicate.
// WARNING	DO NOT USE THIS MACRO unless you fully understand how the alloca() function works,
//		since this macro uses alloca() to allocate memory.  All caveats about using alloca()
//		also apply to this macro.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(char *) == NULL : error, out of memory
//		(char *) != NULL : pointer to duplicate copy of value string
//---------------------------------------------------------------------------------------------------
#define optplus_value_adup(o) ({								\
	optplus_p ___optplus___;								\
	___optplus___ = (o);									\
	char * ___ptr___;									\
	___ptr___ = alloca( ___optplus___->value_len + 1 );					\
	if ( ___ptr___ ) {									\
		memcpy( ___ptr___, ___optplus___->value_str, ___optplus___->value_len );	\
		___ptr___[ ___optplus___->value_len ] = 0;					\
	}											\
	___ptr___;										\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_value_or_arg
// purpose	Get the latest option value string selected by optplus_next() or the argument that
//		follows in the argument list.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(const char *) == NULL : no value available
//		(const char *) != NULL : pointer to value string
//---------------------------------------------------------------------------------------------------
#define optplus_value_or_arg(o) ({								\
	optplus_p ___optplus___;								\
	const char * ___ptr___;									\
	___optplus___ = (o);									\
	___ptr___ = ___optplus___->value_str;							\
	if ( ! ___ptr___ ) ___ptr___ = optplus_arg( ___optplus___ );				\
	___ptr___;										\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix
// purpose	Get the latest option prefix string selected by optplus_next().
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(char *) string with prefix of current option
//---------------------------------------------------------------------------------------------------
#define optplus_prefix(o) ((char *)((o)->pre_char))

//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_one
// purpose	Get the first prefix character code.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) first prefix character code.
//---------------------------------------------------------------------------------------------------
#define optplus_prefix_one(o) ((int)((o)->pre_char[0]))

//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_two
// purpose	Get the second prefix character code.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) first prefix character code.
//---------------------------------------------------------------------------------------------------
#define optplus_prefix_two(o) ((int)((o)->pre_char[1]))

//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_code
// purpose	Get the one or two character prefix combined character code.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(unsigned int) prefix combined character code
//---------------------------------------------------------------------------------------------------
#define optplus_prefix_code(o) ({								\
	optplus_p ___optplus___;								\
	___optplus___ = (o);									\
	char_code_2( ___optplus___->pre_char[1], ___optplus___->pre_char[0] );			\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_char
// purpose	Get the latest single option letter selected by optplus_next().
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) == 0 : no option or option is not a single character
//		(int)  > 0 : option letter character code
//---------------------------------------------------------------------------------------------------
#define optplus_char(o) ((int)((o)->char_code[0]))

//---------------------------------------------------------------------------------------------------
// macro	optplus_arg_num
// purpose	Return the argument index (0 for first) of the recent argument
//		returned by a call to optplus_arg().
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) index of current option (0 for first argument)
//---------------------------------------------------------------------------------------------------
#define optplus_arg_num(o) ({									\
	optplus_p ___optplus___;								\
	___optplus___ = (o);									\
	(int) ( ___optplus___->argt - ___optplus___->arga );					\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_short
// purpose	Yield true if the optplus object has a current option with a short prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a short option, 1: is a short option
//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_long
// purpose	Yield true if the optplus object has a current option with a long prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a long option, 1: is a long option
//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_valid
// purpose	Yield true if the optplus object has a current option with either prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a valid option, 1: is a valid option
//---------------------------------------------------------------------------------------------------
#define optplus_prefix_is_short(o) ({								\
	int ___type___;										\
	___type___ = (o)->type;									\
	___type___ > 0 && ___type___ > char_code_2(1,0);					\
})
#define optplus_prefix_is_long(o) ((o)->type >= char_code_2(1,0))
#define optplus_prefix_is_valid(o) ((o)->type>0)

//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_minus
// purpose	Yield true if the optplus object has a current option with a '-' prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a '-' prefix, 1: is a '-' prefix
//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_plus
// purpose	Yield true if the optplus object has a current option with a '+' prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a '+' prefix, 1: is a '+' prefix
//---------------------------------------------------------------------------------------------------
#define optplus_prefix_is_minus(o) ((o)->type=='-')
#define optplus_prefix_is_plus(o) ((o)->type=='+')

//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_minus_minus
// purpose	Yield true if the optplus object has a current option with a "--" prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a "--" prefix, 1: is a "--" prefix
//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_minus_plus
// purpose	Yield true if the optplus object has a current option with a "-+" prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a "-+" prefix, 1: is a "-+" prefix
//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_plus_minus
// purpose	Yield true if the optplus object has a current option with a "+-" prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a "+-" prefix, 1: is a "+-" prefix
//---------------------------------------------------------------------------------------------------
// macro	optplus_prefix_is_plus_plus
// purpose	Yield true if the optplus object has a current option with a "++" prefix.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) 0: is not a "++" prefix, 1: is a "++" prefix
//---------------------------------------------------------------------------------------------------
#define optplus_prefix_is_minus_minus(o) ((o)->type==char_code_2('-','-'))
#define optplus_prefix_is_minus_plus(o) ((o)->type==char_code_2('-','+'))
#define optplus_prefix_is_plus_minus(o) ((o)->type==char_code_2('+','-'))
#define optplus_prefix_is_plus_plus(o) ((o)->type==char_code_2('+','+'))

//---------------------------------------------------------------------------------------------------
// macro	optplus_dup
// purpose	Make a duplicate terminated string from start and end pointers.
// arguments	1 (const char *) start pointer
//		2 (const char *) end pointer
// yields	(char *) pointer to allocated, copied, and terminated string
//---------------------------------------------------------------------------------------------------
#define optplus_dup(s,e) ({									\
	char * ___ptr___;									\
	char * ___str___;									\
	size_t ___len___;									\
	___str___ = (s);									\
	___len___ = (e) - ___str___;								\
	___ptr___ = malloc( ___len___ + 1 );							\
	if ( ___ptr___ ) {									\
		memcpy( ___ptr___, ___str___, ___len___ );					\
		___ptr___[ ___len___ ] = 0;							\
	}											\
	___ptr___;										\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_dupa
// purpose	Make a duplicate terminated string from start and end pointers in automatic storage.
// WARNING	DO NOT USE THIS MACRO unless you fully understand how the alloca() function works,
//		since this macro uses alloca() to allocate memory.  All caveats about using alloca()
//		also apply to this macro.
// arguments	1 (const char *) start pointer
//		2 (const char *) end pointer
// yields	(char *) pointer to allocated, copied, and terminated string
//---------------------------------------------------------------------------------------------------
#define optplus_dupa(s,e) ({									\
	char * ___ptr___;									\
	char * ___str___;									\
	size_t ___len___;									\
	___str___ = (s);									\
	___len___ = (e) - ___str___;								\
	___ptr___ = alloca( ___len___ + 1 );							\
	if ( ___ptr___ ) {									\
		memcpy( ___ptr___, ___str___, ___len___ );					\
		___ptr___[ ___len___ ] = 0;							\
	}											\
	___ptr___;										\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_original_argc
// purpose	Obtain the argc value for the original arguments.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) argc value for original arguments
//---------------------------------------------------------------------------------------------------
#define optplus_original_argc(o) ({								\
	optplus_p ___optplus___;								\
	___optplus___ = (o);									\
	(int) ( ___optplus___->argz - ___optplus___->arga );					\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_original_argv
// purpose	Obtain the argv value for the original arguments.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(char **) argv value for original arguments
//---------------------------------------------------------------------------------------------------
#define optplus_original_argv(o) ((char**)((o)->arga))

//---------------------------------------------------------------------------------------------------
// macro	optplus_remaining_argc
// purpose	Obtain the argc value for the remaining arguments.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(int) argc value for remaining arguments
//---------------------------------------------------------------------------------------------------
#define optplus_remaining_argc(o) ({								\
	optplus_p ___optplus___;								\
	___optplus___ = (o);									\
	(int) ( ___optplus___->argz - ___optplus___->argx );					\
})

//---------------------------------------------------------------------------------------------------
// macro	optplus_remaining_argv
// purpose	Obtain the argv value for the remaining arguments.
// arguments	1 (optplus_p) pointer to option scanning object
// yields	(char **) argv value for remaining arguments
//---------------------------------------------------------------------------------------------------
#define optplus_remaining_argv(o) ((char**)((o)->argx))

//---------------------------------------------------------------------------------------------------
// macro	optplus_restore_argc_argv
// purpose	Restore the original arcg and argv values into the given
//		variable names.
// arguments	1 (optplus_p) pointer to option scanning object
//		2 (lvalue int) where to restore the argc part
//		3 (lvalue char **) where to restore the argv part
// yields	(int) number of original args (e.g. value of argc)
//---------------------------------------------------------------------------------------------------
#define optplus_restore_argc_argv(o,c,v) ({							\
	optplus_p ___optplus___;								\
	___optplus____ = (o);									\
	(int) ( (c) = ( ___optplus___->argz - ( (v) = ___optplus___->arga ) ) );		\
})

__DEFINE_END__

__SUFFIX_BEGIN__
#endif /* __OPTPLUS_H__ */
__SUFFIX_END__
