//-----------------------------------------------------------------------------
// Copyright  2003 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/string
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include "string_lib.h"

__PROTO_BEGIN__
//-----------------------------------------------------------------------------
// function	str_app_ipv4_min
//
// purpose	Append to the first string (given with maximum available space)
//		a string of dotted decimal digits representing the IPv4 network
//		address given as an unsigned long, with a specified minimum
//		number of digits per octet.
//
//		Do append only if all the digits can be appended.
//
// arguments	1 (char *) pointer to target string
//		2 (size_t) maximum space in target string
//		3 (unsigned long) the IPv4 address to convert
//		4 (int) minimum number of digits per octet.
//
// returns	(size_t) new length of string, or ~0 for error
//
// note		Contents of available space after the termination character
//		are destroyed if the append fails due to insufficient space.
//		Nothing should be kept there, anyway.
//-----------------------------------------------------------------------------
size_t
str_app_ipv4_min (
    char *		arg_target
    ,
    size_t		arg_length
    ,
    unsigned long	arg_ipv4
    ,
    int			arg_min
    )
__PROTO_END__
{
    char *		target_ptr		;
    char *		target_end			;
    size_t		target_len		;

    //-- Determine end of string and check for space.
    target_end = ( target_ptr = arg_target ) + ( target_len = arg_length ) - 1;
    while ( target_ptr < target_end && * target_ptr ) ++ target_ptr;
    if ( target_ptr >= target_end ) return ~0;
    target_len -= target_ptr - arg_target;

    //-- Break address into parts and convert separately.
    if ( str_app_ui_dec( target_ptr, target_len, ( arg_ipv4 >> 24 ) & 0xff, arg_min, '0' ) == ~0 ) goto fail;
    if ( str_app_ch( target_ptr, target_len, '.' ) == ~0 ) goto fail;
    if ( str_app_ui_dec( target_ptr, target_len, ( arg_ipv4 >> 16 ) & 0xff, arg_min, '0' ) == ~0 ) goto fail;
    if ( str_app_ch( target_ptr, target_len, '.' ) == ~0 ) goto fail;
    if ( str_app_ui_dec( target_ptr, target_len, ( arg_ipv4 >>  8 ) & 0xff, arg_min, '0' ) == ~0 ) goto fail;
    if ( str_app_ch( target_ptr, target_len, '.' ) == ~0 ) goto fail;
    if ( str_app_ui_dec( target_ptr, target_len, ( arg_ipv4       ) & 0xff, arg_min, '0' ) == ~0 ) goto fail;

    //-- Get new length and return it.
    while ( target_ptr < target_end && * target_ptr ) ++ target_ptr;
    return target_ptr - arg_target;

 fail:
    //-- Undo appends, restore original string, and return error.
    * target_ptr = 0;
    return ~0;
}
