//-----------------------------------------------------------------------------
// Copyright  2003 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/string
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include "string_lib.h"

__PROTO_BEGIN__
//-----------------------------------------------------------------------------
// function	str_app_rep
//
// purpose	Append to the target string the contents of the source string,
//		with all occurrences of the pattern string replaced by the
//		replacement string.
//
//		If entire resultant string will not fit the available space,
//		then rollback everything leaving the original target string.
//
// arguments	1 (char *) pointer to beginning of target string
//		2 (size_t) total maximum length of target string
//		3 (const char *) pattern string to be matched in source
//		4 (const char *) replacement string to replace each pattern
//		5 (const char *) source string to be appended to target
//
// returns	(size_t) final length of target string
//		(size_t) ~0 : error
//
// note		Contents of target string space located beyond the original
//		string termination character will be overwritten if there is
//		insufficient space to complete the append.  Nothing should be
//		stored there, anyway.
//-----------------------------------------------------------------------------
size_t
str_app_rep (
    char *		arg_target
    ,
    size_t		arg_length
    ,
    const char *	arg_pat
    ,
    const char *	arg_rep
    ,
    const char *	arg_source
    )
__PROTO_END__
{
    char *			target_ptr	;
    char *			target_end	;
    char *			target_undo	;
    const char *		source_ptr	;
    const char *		pat_ptr		;
    const char *		tmp_ptr		;
    int				pat_ch		;

    //---------------------------------------------
    // Determine end of string and check for space.
    //---------------------------------------------
    target_ptr = arg_target;
    target_end = target_ptr + arg_length - 1;
    while ( target_ptr < target_end && * target_ptr ) ++ target_ptr;
    if ( target_ptr >= target_end ) return ~0;
    target_undo = target_ptr;

    //---------------------------------------------------------------
    // Append source string while replacing pattern with replacement.
    //---------------------------------------------------------------
    pat_ch = * arg_pat;
    source_ptr = arg_source;
    while ( * source_ptr ) {
	if ( * source_ptr == pat_ch ) {
	    tmp_ptr = source_ptr;
	    pat_ptr = arg_pat;
	    while ( * pat_ptr && * tmp_ptr && ( * pat_ptr == * tmp_ptr ) ) {
		++ pat_ptr;
		++ tmp_ptr;
	    }
	    if ( ! * pat_ptr ) {
		source_ptr = tmp_ptr;
		tmp_ptr = arg_rep;
		while ( * tmp_ptr && ( target_ptr < target_end ) ) {
		    * target_ptr ++ = * tmp_ptr ++;
		}
	    } else {
		* target_ptr ++ = * source_ptr ++;
	    }
	} else {
	    * target_ptr ++ = * source_ptr ++;
	}
    }

    //-------------------------------------------
    // If the string did not fit, undo and abort.
    //-------------------------------------------
    if ( target_ptr >= target_end ) {
	* target_undo = 0;
	return ~0;
    }

    //--------------------------------------------------------
    // The string did fit, so terminate and return new length.
    //--------------------------------------------------------
    * target_ptr = 0;
    return target_ptr - arg_target;
}
