//-----------------------------------------------------------------------------
// Copyright  2003 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/string
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include "string_lib.h"

__PROTO_BEGIN__
//-----------------------------------------------------------------------------
// function	str_app_url
//
// purpose	Append to the first string given with maximum available space
//		a URL formed from several component strings.
//
//		Do append only if all source strings can be appended.
//
// arguments	1 (char *) pointer to target string
//		2 (size_t) maximum space in target string
//		3 (const char *) protocol, "http", "ftp"
//		4 (const char *) username
//		5 (const char *) password
//		6 (const char *) hostname
//		7 (const char *) portnum
//		8 (const char *) fullpath
//		9 (const char *) querystring
//
// returns	(size_t) new length of string, or ~0 for error
//
// note		Contents of available space after the termination character
//		are destroyed if the append fails due to insufficient space.
//		Nothing should be kept there, anyway.
//-----------------------------------------------------------------------------
size_t
str_app_url (
    char *		arg_target
    ,
    size_t		arg_length
    ,
    const char *	arg_protocol
    ,
    const char *	arg_username
    ,
    const char *	arg_password
    ,
    const char *	arg_hostname
    ,
    const char *	arg_portnum
    ,
    const char *	arg_fullpath
    ,
    const char *	arg_querystring
    )
__PROTO_END__
{
    char *		target_ptr	;
    char *		target_end	;
    char *		target_undo	;

    //-- Determine end of string and check for space.
    target_ptr = arg_target;
    target_end = target_ptr + arg_length;
    while ( target_ptr < target_end && * target_ptr ) ++ target_ptr;
    if ( target_ptr >= target_end ) return ~0;

    //-- Remember where to undo the append.
    target_undo = target_ptr;

    //-- Append each URL component.

#define APPEND_CH(t,e,c)			\
	{					\
	    if ( (t) < (e) ) {			\
		* (t) = (c);			\
		++ (t);				\
	    }					\
	}					\

#define APPEND_STR(t,e,s)			\
	{					\
	    register const char * p;		\
	    p = (s);				\
	    while ( (t) < (e) && * p ) {	\
		* (t) = * p;			\
		++ (t);				\
		++ (p);				\
	    }					\
	}					\

    if ( ! arg_protocol ) arg_protocol = "http";
    APPEND_STR( target_ptr, target_end, arg_protocol );
    APPEND_CH( target_ptr, target_end, ':' );
    if ( arg_username || arg_password || arg_hostname || arg_portnum ) {
	APPEND_CH( target_ptr, target_end, '/' );
	APPEND_CH( target_ptr, target_end, '/' );
	if ( arg_username || arg_password ) {
	    if ( arg_username ) APPEND_STR( target_ptr, target_end, arg_username );
	    if ( arg_password ) {
		APPEND_CH( target_ptr, target_end, ':' );
		APPEND_STR( target_ptr, target_end, arg_password );
	    }
	    APPEND_CH( target_ptr, target_end, '@' );
	}
	if ( ! arg_hostname ) arg_hostname = "127.0.0.1";
	APPEND_STR( target_ptr, target_end, arg_hostname );
	if ( arg_portnum ) {
	    APPEND_CH( target_ptr, target_end, ':' );
	    APPEND_STR( target_ptr, target_end, arg_portnum );
	}
    }
    APPEND_CH( target_ptr, target_end, '/' );
    if ( arg_fullpath ) {
	if ( * arg_fullpath == '/' ) ++ arg_fullpath;
	APPEND_STR( target_ptr, target_end, arg_fullpath );
    }
    if ( arg_querystring ) {
	APPEND_CH( target_ptr, target_end, '?' );
	APPEND_STR( target_ptr, target_end, arg_querystring );
    }

    if ( target_ptr == target_end ) {
	* target_undo = 0;
	return ~ (size_t) 0;
    }
    * target_ptr = 0;
    return target_ptr - arg_target;
}
