//-----------------------------------------------------------------------------
// Copyright  2003 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/string
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include "string_lib.h"

__PROTO_BEGIN__
//-----------------------------------------------------------------------------
// function	str_dup_arr_sep
//
// purpose	Duplicate into allocated space the contents of zero or more
//		(multiple) strings provided as an array of string pointers
//		with a separator string inserted between multiple strings.
//
// arguments	1 (const char *) separator string to insert
//		2 (const char * const *) source strings to duplicate
//
// returns	(char *) new allocated duplicate string, caller must free
//		(char *) NULL : error allocating string or other error
//
// warning	If other threads modify any source strings during duplication,
//		then results are unpredictable.
//-----------------------------------------------------------------------------
char *
str_dup_arr_sep (
    const char *		arg_sep
    ,
    const char * const *	arg_source
    )
__PROTO_END__
{
    char *		ptr_string	;

    //-- Determine total length and allocate new space.
    {
	const char * const *	ptr_array	;
	size_t			total_space	;
	size_t			sep_space	;

	//-- Determine the separator space just once.
	{
	    const char *	ptr_sep		;

	    if ( ( ptr_sep = arg_sep ) ) {
		while ( * ptr_sep ) ++ ptr_sep;
		sep_space = ptr_sep - arg_sep;
	    } else {
		sep_space = 0;
	    }
	}

	total_space = 1;
	if ( ( ptr_array = arg_source ) ) {
	    while ( * ptr_array ) {
		const char *	ptr_source	;

		if ( ptr_array > arg_source ) {
		    total_space += sep_space;
		}
		ptr_source = * ptr_array;
		while ( * ptr_source ) ++ ptr_source;
		total_space += ptr_source - * ptr_array;
		++ ptr_array;
	    }
	}
	ptr_string = (char *) malloc( total_space );
	if ( ! ptr_string ) return ptr_string;
    }

    //-- Copy contents to new string.
    {
	const char * const *	ptr_array	;
	char *			ptr_target	;

	ptr_target = ptr_string;
	if ( ( ptr_array = arg_source ) ) {
	    while ( * ptr_array ) {

		//-- If not the first string, include the separator.
		if ( ptr_array > arg_source ) {
		    const char *	ptr_sep		;

		    if ( ( ptr_sep = arg_sep ) ) {
			while ( ( * ptr_target = * ptr_sep ) ) {
			    ++ ptr_target;
			    ++ ptr_sep;
			}
		    }
		}

		//-- Include the given string.
		{
		    const char *	ptr_source	;
		    ptr_source = * ptr_array;
		    while ( ( * ptr_target = * ptr_source ) ) {
			++ ptr_target;
			++ ptr_source;
		    }
		}
		++ ptr_array;
	    }
	}
	* ptr_target = 0;
    }

    //-- Return the new string.
    return ptr_string;
}
