//-----------------------------------------------------------------------------
// Copyright  2004 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/time
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------
// reference	Explanatory Supplement to the Astronomical Almanac,
//		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]
//		http://shop.bn.com/bookSearch/isbnInquiry.asp?isbn=0935702687
//-----------------------------------------------------------------------------

#include <ctype.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libh/io.h>
#include <libh/time.h>
#include <libh/string.h>


//-----------------------------------------------------------------------------
// program	datecalc
//
// purpose	Perform simple calculations on dates and number of days.
//
//		If two dates are given, calculate the number of days between
//		them and output the number in decimal.  The 2nd date is
//		subtracted from the first date so the results will be:
//
//		date1 <  date2 --> negative number of days
//		date1 == date2 --> zero
//		date1 >  date2 --> positive number of days
//
//		If one date is given, use the current date as the 2nd date
//		and calculate the difference as above.
//
//		The keywords "today" or "now" may be used to force today's
//		date to be used as the first date instead of defaulting to
//		being the 2nd date.
//
//		If one date and one or more numbers is given, calculate a
//		new date by adding those number of days to that date.
//
//		If no date and one or more numbers is given, use the current
//		date and calculate as above.
//
//		More than two dates given is an error.
//-----------------------------------------------------------------------------
int
main (
    int		argc
    ,
    char * *	argv
)
{
    char *		end_ptr			;
    char *		pgm_name		;

    const char *	var_prefix		;
    const char *	var_suffix		;

    eday_t		current_date		;
    eday_t		this_date		;
    eday_t		date_list	[2]	;

    long long		days_total		;
    long long		this_num		;

    int			arg_num			;
    int			date_count		;
    int			error_count		;
    int			num_count		;
    int			shell_type		;
    int			env_var			;

    char		tmp_str		[24]	;


    //-----------------------------
    // Do all initialization first.
    //-----------------------------
    pgm_name = str_last_part( argv[0], '/' );

    arg_num = 0;
    days_total = 0;
    date_count = 0;
    num_count = 0;
    error_count = 0;
    env_var = 0;

    shell_type = -1;

    var_prefix = "";
    var_suffix = "";

    //--------------------------------------
    // The default date is the current date.
    //--------------------------------------
    current_date = time_to_eday( time( NULL ) );
    date_list[ 0 ] = current_date;

    //----------------------------
    // Loop through all arguments.
    //----------------------------
    while ( -- argc > 0 ) {
	++ argv;
	++ arg_num;

	//------------------------------------------
	// Check if this is a shell variable option.
	//------------------------------------------
	if ( str_find_arg( argv[0],
			   "-b",
			   "-s",
			   "--bourne",
			   "--bash",
			   "--ksh",
			   "--korn",
			   "--sh",
			   "--shell" ) ) {
	    shell_type = 0;
	    continue;
	}

	//------------------------------------------
	// Check if this is a shell variable option.
	//------------------------------------------
	if ( str_find_arg( argv[0],
			   "-c",
			   "-t",
			   "--csh",
			   "--tcsh" ) ) {
	    shell_type = 1;
	    continue;
	}

	//----------------------------------------------
	// Check if this is for an environment variable.
	//----------------------------------------------
	if ( str_find_arg( argv[0],
			   "-e",
			   "--env" ) ) {
	    env_var = 1;
	    continue;
	}

	//------------------------------------
	// Check for a variable prefix option.
	//------------------------------------
	if ( memcmp( argv[0], "--prefix", 8 ) == 0 ) {
	    if ( argv[0][8] == 0 ) {
		if ( -- argc == 0 ) {
		    fprintf( stderr, "%s: Ran out of arguments after %d: %s\n", pgm_name, arg_num, argv[0] );
		    ++ error_count;
		    break;
		}
		++ arg_num;
		++ argv;
		var_prefix = argv[0];
	    } else if ( argv[0][8] == '=' ) {
		var_prefix = argv[0] + 9;
	    } else {
		fprintf( stderr, "%s: Invalid option %d: %s\n", pgm_name, arg_num, argv[0] );
		++ error_count;
	    }
	    continue;
	}

	//------------------------------------
	// Check for a variable suffix option.
	//------------------------------------
	if ( memcmp( argv[0], "--suffix", 8 ) == 0 ) {
	    if ( argv[0][8] == 0 ) {
		if ( -- argc == 0 ) {
		    fprintf( stderr, "%s: Ran out of arguments after %d: %s\n", pgm_name, arg_num, argv[0] );
		    ++ error_count;
		    break;
		}
		++ arg_num;
		++ argv;
		var_suffix = argv[0];
	    } else if ( argv[0][8] == '=' ) {
		var_suffix = argv[0] + 9;
	    } else {
		fprintf( stderr, "%s: Invalid option %d: %s\n", pgm_name, arg_num, argv[0] );
		++ error_count;
	    }
	    continue;
	}

	//-----------------------------------------------------------
	// If the argument is "today" or "now", use the current time.
	//-----------------------------------------------------------
	if ( str_find_nocase_arg( argv[0], "today", "now" ) ) {
	    date_list[ date_count ] = current_date;
	    ++ date_count;
	    continue;
	}

	//--------------------------------------------------------
	// If the argument is a valid number, add it to the total.
	//--------------------------------------------------------
	this_num = strtoll( argv[0], & end_ptr, 0 );
	if ( ! * end_ptr ) {
	    days_total += this_num;
	    ++ num_count;
	    continue;
	}

	//----------------------------
	// Otherwise convert the date.
	//----------------------------
	this_date = str_gregorian_to_eday( argv[0] );

	//--------------------------------------
	// If the date is bad, report the error.
	//--------------------------------------
	if ( this_date == EDAY_ERROR ) {
	    fprintf( stderr, "%s: Argument %d is not a number or a date: %s\n", pgm_name, arg_num, argv[0] );
	    ++ error_count;
	    continue;
	}

	//--------------------------------------------------
	// If there are already two dates, report the error.
	//--------------------------------------------------
	if ( date_count >= 2 ) {
	    fprintf( stderr, "%s: Argument %d exceeds two dates: %s\n", pgm_name, arg_num, argv[0] );
	    ++ error_count;
	    continue;
	}

	//-----------------
	// Store this date.
	//-----------------
	date_list[ date_count ] = this_date;
	++ date_count;
    }

    //-----------------------------------------
    // For zero or one date, add numbers to it.
    //-----------------------------------------
    if ( date_count <= 1 ) {

	//-------------------------------
	// If there are errors, quit now.
	//-------------------------------
	if ( error_count > 0 ) return 1;

	//---------------------------------------
	// Calculate date and verify valid range.
	//---------------------------------------
	days_total += date_list[0];
	if ( days_total > (long long) EDAY_MAX || days_total < (long long) EDAY_MIN ) {
	    fprintf( stderr, "%s: Calculated date exceeds conversion range\n", pgm_name );
	    return 1;
	}

	//----------------------------------------------------------------
	// Print the calculated new date according to the variable choice.
	//----------------------------------------------------------------
	if ( shell_type == 0 || shell_type == 1 ) {
	    eday_to_str_cpy( tmp_str, sizeof tmp_str, "%Y", (eday_t) days_total );
	    put_var_shell( stdout, shell_type, env_var, var_prefix, "yyyy", var_suffix, tmp_str );
	    eday_to_str_cpy( tmp_str, sizeof tmp_str, "%m", (eday_t) days_total );
	    put_var_shell( stdout, shell_type, env_var, var_prefix, "mm", var_suffix, tmp_str );
	    eday_to_str_cpy( tmp_str, sizeof tmp_str, "%d", (eday_t) days_total );
	    put_var_shell( stdout, shell_type, env_var, var_prefix, "dd", var_suffix, tmp_str );
	} else {
	    eday_to_printf( "%Y/%m/%d\n", (eday_t) days_total );
	}
    }

    //--------------------------------
    // For two dates, find difference.
    //--------------------------------
    else if ( date_count == 2 ) {

	//----------------------------------------------
	// Verify this is just two dates and no numbers.
	//----------------------------------------------
	if ( num_count > 0 ) {
	    fprintf( stderr, "%s: No numbers allowed with two dates\n", pgm_name );
	    return 1;
	}

	//-------------------------------
	// If there are errors, quit now.
	//-------------------------------
	if ( error_count > 0 ) return 1;

	//------------------------------------
	// Calculate and print the difference.
	//------------------------------------
	days_total = date_list[0];
	days_total -= date_list[1];

	if ( shell_type == 0 || shell_type == 1 ) {
	    snprintf( tmp_str, sizeof tmp_str, "%lld", days_total );
	    put_var_shell( stdout, shell_type, env_var, var_prefix, "diff", var_suffix, tmp_str );
	} else {
	    printf( "%lld\n", days_total );
	}
    }

    //-------------------------------------------
    // If too many dates given, report the error.
    //-------------------------------------------
    else {
	fprintf( stderr, "%s: Too many (%d) dates given.\n", pgm_name, date_count );
	return 1;
    }

    return 0;
}
