//-----------------------------------------------------------------------------
// Copyright  2004 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/time
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

#include "time_lib.h"

#include <libh/string.h>

__PROTO_BEGIN__
//-----------------------------------------------------------------------------
// function	eday_to_str_app
//
// purpose	Append to the first string (given with maximum available space)
//		an Earth Day value in the format specified.
//
// arguments	1 (char *) target string to append to
//		2 (size_t) size limit for target string
//		3 (const char *) format string
//		4 (eday_t) Earth Day value
//
// returns	(size_t) new length of target string or ~0 if append fails
//
// format	The format string is copied with format specifiers replaced
//		with character expressions of the specified date or time
//		component.  The available specifiers are:
//		%g	Specify Gregorian calendar (no output)
//		%j	Specify Julian calendar (no output)
//		%l	Specify lower case conversion (no output)
//		%u	Specify upper case conversion (no output)
//		%n	Specify no case conversion (no output)
//		%E	(1) Earth Time as a decimal number
//		%e	(1) Earth Day as a decimal number
//		%t	(1) System time as a decimal number (e.g. time())
//		%J	(1,0) Julian Date (floating point format)
//		%Y	(04) Year number in the specified calendar
//		%Z	(04) Year number with BC if needed
//		%m	(02) Month number in the specified calendar
//		%d	(02) Day of month in the specified calendar
//		%D	(03) Day of year in the specified calendar
//		%w	(1) Day of week number (0=Sunday, 6=Saturay)
//		%W	(02) Week number in the year (0..53)
//		%A	Day of week name full
//		%a	Day of week name abbreviated
//		%B	Month of year name full
//		%b	Month of year name abbreviated
//		%H	(02) Hours (00..23) of day
//		%M	(02) Minutes (00..59) of hour
//		%S	(02) Seconds (00..59) of minute
//		%F	(03) Fractions of a second (default milliseconds)
//		%U	(06) Fractions of a second (default microseconds)
//		%N	(09) Fractions of a second (default nanoseconds)
//		%s	(05) Seconds (00000..86401) of day
//		%L	(06) Lilian day (LDAY, IBM Y2K hack)
//		%X	(06) eXchange day (XDAY, Bob Bemer Y2K hack)
//		%I	(02) Hour (12,01..11) of meridian
//		%p	Meridian hour designator, AM or PM
//		%T	short form for "%H:%M:%S"
//		%r	short form for "%I:%M:%S %p"
//		%%	Literal "%" character
//
// note		Although time format specifiers are valid when formatting
//		an Earth Day value, they are always converted from 0 which
//		represents the beginning of the given day.
//
// note		Although many specifiers are the same as those defined for
//		use in strftime(), many others are different.
//
// modifiers	A sequence of decimal digits may modify the minimum size
//		and leading character for fields.  Some default to a normal
//		size and this may be overridden.  For example, the day of
//		month defaults to a size specifier of "02" if none is given.
//		To force a date to have 1 digit when less than 10, use "1"
//		as the modifier like "%1d".
//-----------------------------------------------------------------------------
size_t
eday_to_str_app (
    char *		arg_target
    ,
    size_t		arg_space
    ,
    const char *	arg_format
    ,
    eday_t		arg_eday
    )
__PROTO_END__
{
    //-- This points to function to append with appropriate case conversion.
    size_t ( *		app_fun		)( char *, size_t, const char * );

    ecal_t		this_ecal	;
    struct tm		this_tm		;

    const char *	format_ptr	;
    char *		app_reset	;

    int *		spec_con	;
    int *		spec_ptr	;

    size_t		app_len		;

    int			spec_one	;
    int			spec_two	;

    int			lead_dig	;
    int			ch		;


    //-- Note where the append string ends now.
    app_reset = arg_target;
    while ( * app_reset ) ++ app_reset;
    app_len = app_reset - arg_target;

    //-- Start with the Gregorian calendar.
    eday_to_ecal_gregorian( arg_eday, & this_ecal );
    ecal_to_tm( & this_ecal, & this_tm );

    //-- Start with no case conversion.
    app_fun = str_app;

    //------------------------------------------------------------------
    // Main loop to parse the format string.
    //------------------------------------------------------------------
    spec_ptr	= NULL;
    lead_dig	= ' ';
    for ( format_ptr = arg_format; ( ch = * format_ptr ); ++ format_ptr ) {

	//-------------------------------------------------------
	// Continue a working specifier.
	//-------------------------------------------------------
	if ( spec_ptr ) {
	    spec_con = NULL;

	    switch ( ch ) {

		//-------------------------------------------------------
		// Collect leading digit and length specifier.
		//-------------------------------------------------------
	    case '0':
		if ( * spec_ptr == 0 ) lead_dig = ch;
		else * spec_ptr *= 10;
		spec_con = spec_ptr;
		break;

	    case '1': case '2': case '3':
	    case '4': case '5': case '6':
	    case '7': case '8': case '9':
		* spec_ptr *= 10;
		* spec_ptr += ( ch - '0' );
		spec_con = spec_ptr;
		break;

		//-------------------------------------------------------
		// Switch to 2nd length specifier.
		//-------------------------------------------------------
	    case '.':
		spec_con = & spec_two;
		spec_two = 0;
		break;

		//-------------------------------------------------------
		// Select Gregorian calendar and recalculate.
		//-------------------------------------------------------
	    case 'g':
		eday_to_ecal_gregorian( arg_eday, & this_ecal );
		ecal_to_tm( & this_ecal, & this_tm );
		break;

		//-------------------------------------------------------
		// Select Julian calendar and recalculate.
		//-------------------------------------------------------
	    case 'j':
		eday_to_ecal_julian( arg_eday, & this_ecal );
		ecal_to_tm( & this_ecal, & this_tm );
		break;

		//-------------------------------------------------------
		// Select upper case conversion.
		//-------------------------------------------------------
	    case 'u':
		app_fun = str_app_upper;
		break;

		//-------------------------------------------------------
		// Select lower case conversion.
		//-------------------------------------------------------
	    case 'l':
		app_fun = str_app_lower;
		break;

		//-------------------------------------------------------
		// Select no case conversion.
		//-------------------------------------------------------
	    case 'n':
		app_fun = str_app;
		break;

		//-------------------------------------------------------
		// Earth Time
		//-------------------------------------------------------
	    case 'E':
		app_len = str_app_sll_dec( arg_target, arg_space, eday_to_etime( arg_eday ), spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Earth Day
		//-------------------------------------------------------
	    case 'e':
		app_len = str_app_sl_dec( arg_target, arg_space, arg_eday, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// System time
		//-------------------------------------------------------
	    case 't':
		app_len = str_app_sl_dec( arg_target, arg_space, eday_to_time( arg_eday ), spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Julian Date (astronomical)
		//-------------------------------------------------------
	    case 'J':
	    {
		char	convert_here	[ 208 ];

		if ( spec_one <= 0 ) spec_one = 1;
		if ( spec_one > 100 ) spec_one = 100;
		if ( spec_two < 0 ) spec_two = 0;
		if ( spec_two > 100 ) spec_two = 100;
		if ( ( spec_ptr == & spec_two ) ?
		     snprintf( convert_here,
			       sizeof (convert_here) - 2,
			       "%*.*f",
			       spec_one,
			       spec_two,
			       eday_to_jd( arg_eday )
		     ) :
		     snprintf( convert_here,
			       sizeof (convert_here) - 2,
			       "%*f",
			       spec_one,
			       eday_to_jd( arg_eday )
		     )
		> 0 ) {
		    app_len = str_app( arg_target, arg_space, convert_here );
		    if ( app_len == ~0 ) goto return_error;
		}
		break;
	    }

		//-------------------------------------------------------
		// Year (0 = 1 BC)
		//-------------------------------------------------------
	    case 'Y':
		if ( spec_one == 0 ) {
		    spec_one = 4;
		    lead_dig = '0';
		}
		app_len = str_app_sl_dec( arg_target, arg_space, this_ecal.year, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

	    case 'Z':
		if ( spec_one == 0 ) {
		    spec_one = 4;
		    lead_dig = '0';
		}
		{
		    long this_year;
		    this_year = this_ecal.year;
		    if ( this_year < 1 ) this_year = 1 - this_year;
		    app_len = str_app_sl_dec( arg_target, arg_space, this_year, spec_one, lead_dig );
		}
		if ( this_ecal.year < 1 ) str_app( arg_target, arg_space, " BC" );
		if ( app_len == ~0 ) goto return_error;
		break;

            case 'y':
                if ( spec_one == 0 ) {
                    spec_one = 2;
                    lead_dig = '0';
                }
                app_len = str_app_sl_dec( arg_target, arg_space, this_ecal.year % 100U, spec_one, lead_dig );
                if ( app_len == ~0 ) goto return_error;
                break;

		//-------------------------------------------------------
		// Month of year
		//-------------------------------------------------------
	    case 'm':
		if ( spec_one == 0 ) {
		    spec_one = 2;
		    lead_dig = '0';
		}
		app_len = str_app_si_dec( arg_target, arg_space, this_ecal.month, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Day of month
		//-------------------------------------------------------
	    case 'd':
		if ( spec_one == 0 ) {
		    spec_one = 2;
		    lead_dig = '0';
		}
		app_len = str_app_ui_dec( arg_target, arg_space, this_ecal.mday, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Day of year
		//-------------------------------------------------------
	    case 'D':
		if ( spec_one == 0 ) {
		    spec_one = 3;
		    lead_dig = '0';
		}
		app_len = str_app_ui_dec( arg_target, arg_space, this_ecal.yday + 1, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Day of week
		//-------------------------------------------------------
	    case 'w':
		if ( spec_one == 0 ) {
		    spec_one = 1;
		    lead_dig = '0';
		}
		app_len = str_app_ui_dec( arg_target, arg_space, this_ecal.wday, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Week of year
		//-------------------------------------------------------
	    case 'W':
		if ( spec_one == 0 ) {
		    spec_one = 2;
		    lead_dig = '0';
		}
		app_len = str_app_ui_dec( arg_target, arg_space, this_ecal.week, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Abbreviated day of week name
		//-------------------------------------------------------
	    case 'a':
		
		( * app_fun )( arg_target, arg_space, day_abbrev( this_ecal.wday - 1 ) );
		break;

		//-------------------------------------------------------
		// Full day of week name
		//-------------------------------------------------------
	    case 'A':
		( * app_fun )( arg_target, arg_space, day_name( this_ecal.wday - 1 ) );
		break;

		//-------------------------------------------------------
		// Abbreviated month of year name
		//-------------------------------------------------------
	    case 'b':
		( * app_fun )( arg_target, arg_space, month_abbrev( this_ecal.month - 1 ) );
		break;

		//-------------------------------------------------------
		// Full month of year name
		//-------------------------------------------------------
	    case 'B':
		( * app_fun )( arg_target, arg_space, month_name( this_ecal.month - 1 ) );
		break;

		//-------------------------------------------------------
		// Hour of a 12 hour clock.
		//-------------------------------------------------------
	    case 'I':
		if ( spec_one == 0 ) {
		    spec_one = 2;
		    lead_dig = '0';
		}
		app_len = str_app_ui_dec( arg_target, arg_space, 12, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Show the Post-Meridian or Anti-Meridian designator.
		//-------------------------------------------------------
	    case 'p':
		app_len = str_app( arg_target, arg_space, "AM" );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Hour
		//-------------------------------------------------------
	    case 'H':
		if ( spec_one == 0 ) {
		    spec_one = 2;
		    lead_dig = '0';
		}
		app_len = str_app_ui_dec( arg_target, arg_space, 0, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Minute
		//-------------------------------------------------------
	    case 'M':
		if ( spec_one == 0 ) {
		    spec_one = 2;
		    lead_dig = '0';
		}
		app_len = str_app_ui_dec( arg_target, arg_space, 0, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Second
		//-------------------------------------------------------
	    case 'S':
		if ( spec_one == 0 ) {
		    spec_one = 2;
		    lead_dig = '0';
		}
		app_len = str_app_ui_dec( arg_target, arg_space, 0, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Convert like "%H:%M:%S".
		//-------------------------------------------------------
	    case 'T':
		app_len = str_app( arg_target, arg_space, "00:00:00" );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Convert like "%I:%M:%S %p".
		//-------------------------------------------------------
	    case 'r':
		app_len = str_app( arg_target, arg_space, "12:00:00 AM" );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// fractions of a second
		//-------------------------------------------------------
	    case 'f':
	    case 'F':
	    case 'U':
	    case 'N':
		if ( spec_one == 0 ) {
		    spec_one = ch == 'N' ? 9 : ch == 'U' ? 6 : 3;
		    lead_dig = '0';
		}
		{
		    uint64_t	frac_sec	;
		    uint64_t	frac_div	;
		    char *	target_ptr	;
		    char *	target_end	;

		    frac_sec = this_ecal.nsec;
		    frac_sec *= UINT64_C(10000000000);
		    frac_div = UINT64_C(10000000000000000000);

		    target_ptr = arg_target;
		    target_end = arg_target + arg_space - 1;

		    //-- Find end of terminated string.
		    while ( target_ptr < target_end && * target_ptr ) ++ target_ptr;

		    //-- Convert fraction digits from most to least significant.
		    while ( target_ptr < target_end && spec_one > 0 ) {
			if ( frac_div ) {
			    frac_div /= 10U;
			    * target_ptr ++ = "0123456789"[ ( frac_sec / frac_div ) % UINT64_C(10) ];
			} else {
			    * target_ptr ++ = '0';
			}
		    }
		    * target_ptr = 0;
		}
		break;

		//-------------------------------------------------------
		// Second of day
		//-------------------------------------------------------
	    case 's':
		if ( spec_one == 0 ) {
		    spec_one = 5;
		    lead_dig = '0';
		}
		app_len = str_app_ul_dec( arg_target, arg_space, 0, spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Lilian day (LDAY, IBM Y2K hack)
		//-------------------------------------------------------
	    case 'L':
		if ( spec_one == 0 ) {
		    spec_one = 6;
		    lead_dig = '0';
		}
		app_len = str_app_si_dec( arg_target, arg_space, eday_to_lday( arg_eday ), spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// eXchange day (XDAY, Bob Bemer Y2K hack)
		//-------------------------------------------------------
	    case 'X':
		if ( spec_one == 0 ) {
		    spec_one = 6;
		    lead_dig = '0';
		}
		app_len = str_app_si_dec( arg_target, arg_space, eday_to_xday( arg_eday ), spec_one, lead_dig );
		if ( app_len == ~0 ) goto return_error;
		break;

		//-------------------------------------------------------
		// Literal %
		//-------------------------------------------------------
	    case '%':

		//-------------------------------------------------------
		// Copy anything else as is.
		//-------------------------------------------------------
	    default:
		app_len = str_app_ch( arg_target, arg_space, ch );
		if ( app_len == ~0 ) goto return_error;
	    }
	    spec_ptr = spec_con;
	}

	//-------------------------------------------------------
	// Start a format specifier.
	//-------------------------------------------------------
	else if ( ch == '%' ) {
	    spec_ptr = & spec_one;
	    spec_one = 0;
	    lead_dig = ' ';
	}

	//-------------------------------------------------------
	// Copy anything else as is.
	//-------------------------------------------------------
	else {
	    app_len = str_app_ch( arg_target, arg_space, ch );
	    if ( app_len == ~0 ) goto return_error;
	}

    }

    //------------------------------------------
    // Return an allocated string to the caller.
    //------------------------------------------
    return app_len;


 return_error:
    //--------------------------------------------------
    // Restore the append string to the original length.
    //--------------------------------------------------
    * app_reset = 0;
    return app_len;
}
