//-----------------------------------------------------------------------------
// Copyright  2004 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	libh/time
// homepage	http://libh.slashusr.org/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------
// reference	Explanatory Supplement to the Astronomical Almanac,
//		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]
//		http://shop.bn.com/bookSearch/isbnInquiry.asp?isbn=0935702687
//-----------------------------------------------------------------------------

#include <ctype.h>

#include "time_lib.h"


__PROTO_BEGIN__
//-----------------------------------------------------------------------------
// function	ymd_julian_to_eday
//
// purpose	Convert a Julian calendar date to an Earth Day value.
//
// arguments	1 (long) Julian calendar year (0 = 1 BC, 1 = 1 AD)
//		2 (int) Julian calendar month (1-12)
//		3 (int) Julian calendar date (1-31)
//
// returns	(eday_t) resultant Earth Day value
//
// note		The calculation is for the proleptic Julian calendar.
//		There is no accomodation for selection between Julian and
//		Gregorian calendars.  Since different areas of the world
//		switched calendars at different dates, the calling program
//		will have to determine on its own what is the appropriate
//		calendar context and call ymd_gregorian_to_eday() or ymd_julian_to_eday().
//-----------------------------------------------------------------------------
eday_t
ymd_julian_to_eday (
    long	arg_year
    ,
    int		arg_month
    ,
    int		arg_date
    )
__PROTO_END__
{
    //-----------------------------------------------------
    // Define a lookup table indexed by the month offset in
    // a year yielding the number of days to that month.
    //-----------------------------------------------------
    static int month_to_day[12] = {
	0,31,59,90,120,151,181,212,243,273,304,334
    };

    //-----------------------------------------------------
    // Define a lookup table indexed by the month offset in
    // a year yielding the number of days in that month.
    //-----------------------------------------------------
    static int days_in_month[12] = {
	31,28,31,30,31,30,31,31,30,31,30,31
    };

    //------------------------
    // Define other variables.
    //------------------------
    long		cycle		;
    long		days		;
    int			year		;
    int			month		;


    //-----------------------------------------------------------
    // Split year into cycle number and year offset within cycle.
    //-----------------------------------------------------------
    cycle = quo_ni( arg_year, 4 );
    year = arg_year - cycle * 4;

    //------------------------------
    // Get month into range 0 to 11.
    //------------------------------
    month = arg_month - 1;
    if ( arg_month < 1 ) month = 0;
    else if ( arg_month > 12 ) month = 11;

    //------------------------------------
    // Get days into range for this month.
    //------------------------------------
    days = days_in_month[ month ] + ( year == 0 && month == 1 );
    if ( arg_date < 1 ) days = 1;
    else if ( arg_date < days ) days = arg_date;
    -- days;

    //--------------------------------
    // Add days in year to this month.
    //--------------------------------
    days += month_to_day[ month ];
    if ( year == 0 && month > 1 ) ++ days;

    //--------------------------------------------------------
    // Add days in cycle to this year, days to cycle from year
    // zero, adjust to epoch, and return this number of days.
    //--------------------------------------------------------
    return days + ( 1461 * year + 3 ) / 4 + cycle * 1461 + EDAY_1_JAN_0_JULIAN;
}
