//-----------------------------------------------------------------------------
// Copyright  2010 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	libh at ipal dot org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 120 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// program	tcpdelay
//
// purpose	Accept connections and delay them by accepting and discarding
//		input, and outptting nothing.
//
// syntax	tcpdelay  [options]  listenport
//
//		-c		enable capture (in changed directory)
//		-d <dir>	change directory
//		-h		show help
//		-l <file>	log file
//              -p <file>	store master process ID here
//		-q		be quiet with startup messages
//		-u <user>	switch user after binding socket
//		-v		be verbose with startup messages
//		-V		show version number
//
// note		The chroot() system call will be made to change the root to
//		the current directory after all directory changes are done.
//		Combined with changing to a safe directory and switching to
//		a non-root user, this offers some protection against possible
//		exploits, few as they may be given the simplicity of this.
//
// requires	LIBH
//
// compile	gcc -O2 -o tcpdelay tcpdelay.c -lh
//-----------------------------------------------------------------------------
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>


#define _GNU_SOURCE
#include <getopt.h>
extern char *optarg;
extern int optind, opterr, optopt;

#include <libh/daemon.h>
#include <libh/io.h>
#include <libh/net.h>
#include <libh/string.h>
#include <libh/time.h>

extern int h_errno;


//-----------------------------------------------------------------------------
// Globally shared variables.
//-----------------------------------------------------------------------------
static char *		program_name	;


//-----------------------------------------------------------------------------
// function	show_help
//-----------------------------------------------------------------------------
int
show_help ()
{
    fprintf( stderr,
	     "\n"
	     "syntax:  %s  [options]  listenport\n"
	     "\n"
	     "options: -c               enable traffic capture to a file\n"
	     "         -d  <directory>  change directory for files\n"
	     "         -g  <group>      set effective group for files\n"
	     "         -l  <file>       log file\n"
	     "         -h               show this help\n"
	     "         -p  <file>       store master process ID here\n"
	     "         -q               be quite with start messages\n"
	     "         -u  <user>       set effective user for files\n"
	     "         -v               be verbose with start messages\n"
	     "         -V               show version number\n"
	     "\n"
	     "signals: HUP              graceful restart (wait for workers)\n"
	     "         USR1             fast restart (kill workers)\n"
	     "         USR2             graceful shutdown (wait for workers)\n"
	     "         INT,TERM,QUIT    fast shutdown (kill workers)\n"
	     "\n",
	     program_name );
    return 1;
}

//-----------------------------------------------------------------------------
// function	write_capture_data
//
// purpose	Write data to a capture file.
//
// arguments	1 (int) file descriptor
//		2 (int) data type character
//		3 (const char *) data to write
//		4 (size_t) length of data to write
//
// returns	(void)
//
// format	Byte order is big endian.
//		bytes 0-3:	number of seconds since epoch
//		bytes 4-6:	number of microseconds more
//		byte 7:		data chunk type code
//		bytes 8-9:	length of data following header
//-----------------------------------------------------------------------------
static inline
void
write_capture_data (
    int			arg_fd
    ,
    int			arg_type
    ,
    const char *	arg_ptr
    ,
    size_t		arg_len
    )
{
    if ( arg_fd < 0 ) return;

    {
	long long	this_time		;
	char *		ptr			;
	size_t		len			;
	char		header_mem	[10]	;

	this_time = current_timeus();
	ptr = header_mem;
	*    ptr = this_time >> 48;
	* ++ ptr = this_time >> 40;
	* ++ ptr = this_time >> 32;
	* ++ ptr = this_time >> 24;
	* ++ ptr = this_time >> 16;
	* ++ ptr = this_time >>  8;
	* ++ ptr = this_time;
	* ++ ptr = arg_type;
	* ++ ptr = arg_len >> 8;
	* ++ ptr = arg_len;

	len = 10;
	ptr = header_mem;
	while ( len ) {
	    ssize_t write_len;
	    if ( ( write_len = write( arg_fd, ptr, len ) ) < 0 ) return;
	    ptr += write_len;
	    len -= write_len;
	}
    }

    {
	const char *ptr;
	size_t len;

	len = arg_len;
	ptr = arg_ptr;
	while ( len ) {
	    ssize_t write_len;
	    if ( ( write_len = write( arg_fd, ptr, len ) ) < 0 ) return;
	    ptr += write_len;
	    len -= write_len;
	}
    }

    return;
}

//-----------------------------------------------------------------------------
// Main program.
//-----------------------------------------------------------------------------
int
main (
    int		argc
    ,
    char * *	argv
    )
{
    static const struct option	long_opts	[]	= {
	{ "cap",	no_argument,		NULL,	'c' },
	{ "capture",	no_argument,		NULL,	'c' },
	{ "dir",	required_argument,	NULL,	'd' },
	{ "directory",	required_argument,	NULL,	'd' },
	{ "chdir",	required_argument,	NULL,	'd' },
	{ "chroot",	required_argument,	NULL,	'd' },
	{ "gid",	required_argument,	NULL,	'g' },
	{ "group",	required_argument,	NULL,	'g' },
	{ "help",	no_argument,		NULL,	'h' },
	{ "logfile",	required_argument,	NULL,	'l' },
	{ "log",	required_argument,	NULL,	'l' },
	{ "pidfile",	required_argument,	NULL,	'p' },
	{ "quiet",	no_argument,		NULL,	'q' },
	{ "uid",	required_argument,	NULL,	'u' },
	{ "user",	required_argument,	NULL,	'u' },
	{ "verbose",	no_argument,		NULL,	'v' },
	{ "version",	no_argument,		NULL,	'V' },
	{ NULL,		0,			NULL,	0 }
    };

    static const int		num1			= 1;

    static const char		short_opts	[]	= ":cd:g:hl:p:qu:vV";


    struct sockaddr_in	listen_addr		;
    struct sockaddr_in	client_addr		;

    socklen_t		client_addrlen		;

    char *		buffer_ptr		;

    char *		listen_port_str		;
    char *		opt_group		;
    char *		opt_user		;
    char *		opt_pidfile		;
    char *		opt_end			;

    long		start_count		;

    size_t		buffer_len		;

    int			listen_fd		;
    int			capture_fd		;
    int			listen_port		;
    int			master_pid		;
    int			opt_help		;
    int			opt_capture		;
    int			opt_verbose		;
    int			opt_version		;
    int			error_count		;
    int			error_chdir		;

    gid_t		opt_gid			;
    uid_t		opt_uid			;

    char		dir_str		[PATH_MAX+8]	;
    char		path_str	[PATH_MAX+8]	;


    //----------------------
    // Initialize variables.
    //----------------------
    error_count = 0;
    error_chdir = 0;

    opt_group = NULL;
    opt_user = NULL;
    opt_pidfile = NULL;

    opt_help = 0;
    opt_capture = 0;
    opt_verbose = -1;

    opt_gid = -1;
    opt_uid = -1;

    capture_fd = -1;

    //--------------------------------------------------------
    // At this point main() is running in the startup process.
    // This is the process that the shell is waiting to exit.
    //--------------------------------------------------------
    program_name = str_tail_one( argv[0], '/' );
    program_name = strdup( program_name );
    daemon_set_program_name( program_name );

    //--------------
    // Scan options.
    //--------------
    for (;;) {
	int opt;

	opt = getopt_long( argc, argv, short_opts, long_opts, NULL );
	if ( opt < 0 ) break;

	switch ( opt ) {

	case 'c':
	    opt_capture = 1;
	    continue;

	case 'd':
	    if ( error_chdir == 0 ) {
		if ( chdir( optarg ) < 0 ) {
		    fprintf( stderr, "%s: chdir(\"%s\"): %s\n", program_name, optarg, strerror( errno ) );
		    ++ error_count;
		    ++ error_chdir;
		}
	    }
	    else if ( error_chdir == 1 ) {
		fprintf( stderr, "%s: no further messages will be issued for -d options", program_name );
		++ error_chdir;
	    }
	    continue;

	case 'g':
	    opt_gid = strtoul( optarg, & opt_end, 0 );
	    if ( * opt_end ) {
		struct group * this_group;
		opt_group = optarg;
		this_group = getgrnam( opt_group );
		if ( ! this_group ) {
		    fprintf( stderr, "%s: group \"%s\" not found\n", program_name, opt_group );
		    ++ error_count;
		    continue;
		}
		opt_gid = this_group->gr_gid;
	    }
	    continue;

	case 'h':
	    ++ opt_help;
	    continue;

	case 'p':
	    opt_pidfile = optarg;
	    continue;

	case 'q':
	    opt_verbose = 0;
	    continue;

	case 'u':
	    opt_uid = strtoul( optarg, & opt_end, 0 );
	    if ( * opt_end ) {
		struct passwd * this_user;
		opt_user = optarg;
		this_user = getpwnam( opt_user );
		if ( ! this_user ) {
		    fprintf( stderr, "%s: user \"%s\" not found\n", program_name, opt_user );
		    ++ error_count;
		    continue;
		}
		opt_uid = this_user->pw_uid;
	    }
	    continue;

	case 'v':
	    opt_verbose = 1;
	    continue;

	case 'V':
	    ++ opt_version;
	    continue;

	default:
	    if ( opt ) {
		fprintf( stderr, "%s: unrecognized option '%c'\n", program_name, opt );
	    } else {
		fprintf( stderr, "%s: unrecognized option \"%s\"\n", program_name, argv[optind] );
	    }
	    ++ error_count;
	    continue;
	}
    }

    //----------------------------------
    // Quit now if there are any errors.
    //----------------------------------
    if ( error_count ) return 1;

    //---------------------------
    // One arguments is required.
    //---------------------------
    if ( argc - optind < 1 ) return show_help();
    listen_port_str = argv[ optind + 0 ];

    //--------------------------------------------
    // Convert or look up listen port to a number.
    //--------------------------------------------
    listen_port = get_port_by_name( listen_port_str, "tcp" );
    if ( listen_port < 0 ) {
	fprintf( stderr, "%s: invalid listen port \"%s\"\n", program_name, listen_port_str );
	return 1;
    }

    //-----------------------------------------------
    // Switch from startup process to master process.
    //-----------------------------------------------
    daemon_set_start_timeout( 30 );
    if ( daemon_start() < 0 ) _exit( 1 );
    argv[0][0] = 'M';
    master_pid = getpid();

    //--------------------------
    // Set up the listen socket.
    //--------------------------
    listen_fd = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
    if ( listen_fd < 0 ) {
	fprintf( stderr, "%s: socket(PF_INET,SOCK_STREAM,IPPROTO_TCP): %s\n",
		 program_name, strerror( errno ) );
	_exit( 1 );
    }
    if ( setsockopt( listen_fd, SOL_SOCKET, SO_REUSEADDR, (void*) & num1, (socklen_t) sizeof num1 ) < 0 ) {
	fprintf( stderr, "%s: setsockopt(,SOL_SOCKET,SO_REUSEADDR,,): %s\n",
		 program_name, strerror( errno ) );
	_exit( 1 );
    }
    listen_addr = hbo_to_sockaddr_in( 0, listen_port );
//    listen_addr.sin_family = AF_INET;
//    listen_addr.sin_port = htons( (uint16_t) listen_port );
//    listen_addr.sin_addr.s_addr = 0;
    if ( bind( listen_fd, (struct sockaddr *) & listen_addr, (socklen_t) sizeof listen_addr ) < 0 ) {
	if ( errno == EADDRINUSE ) { // Output a more meaningful error message.
	    fprintf( stderr, "%s: bind: Port %u already in use\n",
		     program_name, listen_port );
	} else {
	    fprintf( stderr, "%s: bind port %u: %s\n",
		     program_name, listen_port, strerror( errno ) );
	}
	_exit( 1 );
    }
    if ( listen( listen_fd, 128 ) < 0 ) {
	fprintf( stderr, "%s: listen(%d,128): %s\n",
		 program_name, listen_fd, strerror( errno ) );
	_exit( 1 );
    }

    //-------------------------------------
    // Write master process ID to pid file.
    //-------------------------------------
    if ( opt_pidfile ) {
	FILE * pidfile;
	char pidpath[ PATH_MAX + 1 ];
	snprintf( pidpath, sizeof pidpath, "%s.new", opt_pidfile );
	pidfile = fopen( pidpath, "w" );
	if ( pidfile ) {
	    fprintf( pidfile, "%u\n", master_pid );
	    fclose( pidfile );
	    rename( pidpath, opt_pidfile );
	}
    }

    //--------------------------------------------------------------
    // Note what directory we are now in before chroot() hides this.
    //--------------------------------------------------------------
    if ( ! getcwd( dir_str, sizeof dir_str ) ) {
        fprintf( stderr, "%s: getcwd( %p, %lu ): %s\n",
		 program_name, dir_str, (unsigned long) sizeof dir_str, strerror( errno ) );
	++ error_count;
    }

    //--------------------------------------------------------------------
    // Try to change this process root directory to the current directory.
    // This will fail if not running as root, but ignore that error.
    //--------------------------------------------------------------------
    if ( chroot( "." ) < 0 && geteuid() == 0 ) {
	fprintf( stderr, "%s: chroot( \".\" ): %s\n",
		 program_name, strerror( errno ) );
    }

    //---------------------------------------
    // Switch user and/or group as specified.
    //---------------------------------------
    if ( opt_gid >= 0 && setregid( opt_gid, opt_gid ) < 0 ) {
	fprintf( stderr, "%s: unable to switch to group id %u (specified as \"%s\"): %s\n",
		 program_name, opt_gid, opt_group, strerror( errno ) );
	++ error_count;
    }
    if ( opt_uid >= 0 && setreuid( opt_uid, opt_uid ) < 0 ) {
	fprintf( stderr, "%s: unable to switch to user id %u (specified as \"%s\"): %s\n",
		 program_name, opt_uid, opt_user, strerror( errno ) );
	++ error_count;
    }
    if ( error_count ) {
	fflush( NULL );
	_exit( 1 );
    }

    //----------------------------------------------
    // Switch from master process to server process.
    //----------------------------------------------
    daemon_set_restart_count_max( 256 );
    if ( daemon_master( & start_count ) != 0 ) {
	fprintf( stderr, "%s: server restart limit (%lu) exceeded\n", program_name, start_count );
	fflush( NULL );
	_exit( 1 );
    }
    argv[0][0] = 'S';

    //------------------------------------------------------------------------
    // If this is the first server start, output a message and signal success.
    //------------------------------------------------------------------------
    if ( start_count == 0 ) {
	if ( opt_verbose ) {
	    printf( "%s[%u]: port %s", program_name, master_pid, listen_port_str );
	    if ( listen_port_str[0] < '0' || '9' < listen_port_str[0] ) {
		printf( "(%u)", listen_port );
	    }
	    fputc( '\n', stdout );

	    if ( opt_capture ) {
		printf( "%s[%u]: port %d logging in directory \"%s\"\n",
			program_name, master_pid, listen_port, dir_str );
	    }

	    fflush( stdout );
	}
	daemon_success();
    }

    //-------------------------------------------------------------------
    // Run server loop which starts a worker process for each connection.
    //-------------------------------------------------------------------
    daemon_set_worker_count_max( 40 );
    {
	int fd_list[2];
	fd_list[0] = listen_fd;
	fd_list[1] = -1;
	client_addrlen = sizeof client_addr;
	if ( daemon_server( fd_list, (struct sockaddr *) & client_addr, & client_addrlen ) < 0 ) _exit( 1 );
    }
    argv[0][0] = 'W';

    //-------------------------------------
    // Allocate a buffer to read data into.
    //-------------------------------------
    buffer_len = 65536;
    buffer_ptr = malloc( buffer_len );
    if ( ! buffer_ptr ) _exit( 1 );

    //---------------------
    // Open a capture file.
    //---------------------
    if ( opt_capture ) {
	size_t len;
	current_strftime_usec_local( path_str, sizeof path_str, "%Y/%m/%d/%H%M%S", "-%06lu" );
	len = strlen( path_str );
	snprintf( path_str + len, sizeof path_str - len, "-%05u", (unsigned int) ( getpid() & 0xffff ) );
	make_parent_dir( path_str, 0700 );
	capture_fd = open( path_str, O_CREAT | O_WRONLY, 0600 );
    }

    //----------------------------------------
    // Capture where the connection came from.
    //----------------------------------------
    if ( capture_fd >= 0 ) {
	char buf[6];
	memcpy( buf + 0, & client_addr.sin_addr.s_addr, 4 );
	memcpy( buf + 4, & client_addr.sin_port, 2 );
	write_capture_data( capture_fd, 'a', buf, 6 );
    }

    //------------------------------------
    // We won't be using the duplicate fd.
    //------------------------------------
    close( 1 );

    //----------------------------
    // This is the main read loop.
    //----------------------------
    for (;;) {
	ssize_t len;
	len = read( 0, buffer_ptr, buffer_len );
	if ( len <= 0 ) break;
	if ( capture_fd >= 0 ) write_capture_data( capture_fd, 's', buffer_ptr, len );
    }
    write_capture_data( capture_fd, 's', NULL, 0 );

    //-------
    // Close.
    //-------
    shutdown( 0, SHUT_RD );
    shutdown( 0, SHUT_WR );
    close( 0 );

    //-----------------------
    // Capture when complete.
    //-----------------------
    if ( capture_fd >= 0 ) {
	write_capture_data( capture_fd, 'z', NULL, 0 );
	close( capture_fd );
    }

    //--------------------
    // Exit child process.
    //--------------------
    _exit( 0 );
}
